from .utils import get_linter_errors_list


def test_e201_triggered(sitefab, empty_post):
    sitefab.plugin_data['image_info'] = {}
    empty_post.elements.images = ["test.jpg"]
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E201" in error_list


def test_e201_not_triggered(sitefab, empty_post):
    sitefab.plugin_data['image_info'] = {
        "test.jpg": {}
    }
    empty_post.elements.images = ["test.jpg"]
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E201" not in error_list


def test_e203_triggered(sitefab, empty_post):
    empty_post.elements.images = ["test.jpg", "test.jpg"]
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E203" in error_list


def test_e203_not_triggered(sitefab, empty_post):
    empty_post.elements.images = ["test.jpg"]
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E203" not in error_list


def test_e204_triggered(sitefab, empty_post):
    sitefab.plugin_data['image_info'] = {
        "banner.jpg": {"width": 100, "height": 34}
    }
    empty_post.meta.banner = "banner.jpg"
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E204" in error_list


def test_e204_not_triggered(sitefab, empty_post):
    sitefab.plugin_data['image_info'] = {
        "banner.jpg": {"width": 1900, "height": 34}
    }
    empty_post.meta.banner = "banner.jpg"
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E204" not in error_list


def test_e205_triggered(sitefab, empty_post):
    sitefab.plugin_data['image_info'] = {
        "test.jpg": {"width": 50, "height": 34}
    }
    empty_post.elements.images = ["test.jpg"]
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E205" in error_list


def test_e205_not_triggered(sitefab, empty_post):
    sitefab.plugin_data['image_info'] = {
        "test.jpg": {"width": 10000, "height": 34}
    }
    empty_post.elements.images = ["test.jpg"]
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E205" not in error_list


def test_e206_triggered(sitefab, empty_post):
    sitefab.plugin_data['image_info'] = {
        "banner.jpg": {"width": 10000, "height": 34}
    }
    empty_post.meta.banner = "banner-typo.jpg"
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E206" in error_list


def test_e206_not_triggered(sitefab, empty_post):
    sitefab.plugin_data['image_info'] = {
        "banner.jpg": {"width": 10000, "height": 34}
    }
    empty_post.meta.banner = "banner.jpg"
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E206" not in error_list


def test_e207_triggered(sitefab, empty_post):
    sitefab.plugin_data['image_info'] = {
        "banner.jpg": {"width": 10000, "height": 34}
    }
    empty_post.meta.banner = "banner.jpg"
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E207" in error_list


def test_e207_not_triggered(sitefab, empty_post):
    sitefab.plugin_data['image_info'] = {
        "banner.jpg": {"width": 1900, "height": 1080}
    }
    empty_post.meta.banner = "banner.jpg"
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E207" not in error_list



import re
import yaml
from yaml import YAMLError
import datetime
import time
from sitefab import utils

date_matcher = re.compile('(\d+) +(\w{3}) +(\d+) +(\d+):(\d+)')  # noqa
frontmatter_matcher = re.compile(r'(^\s*---.*?---\s*$)',
                                 re.DOTALL | re.MULTILINE)


def parse_fields(fields=None):
    """ Recursively parse a given dict of fields to add extra information
    (e.g timestamp) if needed

    Args:
        fields (dict): the fields to parse

    Returns:
        objdict: the fields with the additional properties
    """
    new_fields = {}
    if fields:
        for name, value in fields.items():
            if isinstance(value, dict):
                new_fields[name] = parse_fields(value)
            else:
                new_fields[name] = value
                # adding timestamp
                if value and "_date" in name:
                    ts = parse_date_to_ts(value)
                    if ts:
                        fts = name + "_ts"
                        new_fields[fts] = ts

    return new_fields


def parse_date_to_ts(date_str):
    """ create the timestamp coresponding to a given date string"""
    if not date_str:
        return None
    m = date_matcher.search(date_str)

    if not m:
        return None
    day = m.group(1)
    month = m.group(2).capitalize()
    year = m.group(3)
    hour = m.group(4)
    minutes = m.group(5)

    if len(day) == 1:
        day = "0" + day

    if len(hour) == 1:
        hour = "0" + hour

    if len(minutes) == 1:
        minutes = "0" + minutes
    date_str = "%s-%s-%s %s:%s" % (day, month, year, hour, minutes)
    try:
        d = datetime.datetime.strptime(date_str, "%d-%b-%Y %H:%M")
    except:  # noqa
        return None
    dtt = d.timetuple()  # time.struct_time
    ts = int(time.mktime(dtt))
    ts -= (3600 * 8)
    return ts


def parse(post):
    """ Get a post content and extract frontmatter data if exist

    Args:
        post (str): post to parse

    Returns
        list: [meta data, md]

    note: all sanity check must be done via the linter and
    used in linter.validate()
    """
    md = post
    d = frontmatter_matcher.search(post)
    if d:
        frontmatter = d.group(1)
        md = md.replace(frontmatter, "")
        frontmatter = frontmatter.replace("---", '')
        try:
            m = yaml.load(frontmatter)  # using YAML :)
        except YAMLError as ye:
            print(ye)
            m = None

        if type(m) != dict:
            meta_data = None
        else:
            meta_data = parse_fields(m)
            meta = utils.dict_to_objdict(meta_data)

    return [meta, md]



from sitefab import image


def test_img_extensions():
    values = [
        [".png", "PNG", "image/png"],
        [".webp", "WEBP", "image/webp"],
        [".jpeg", "JPEG", "image/jpeg"],
        [".jpg", "JPEG", "image/jpeg"],
        ["unkn", None, None]
    ]
    for value in values:
        codename, extension = image.normalize_image_extension(value[0])
        assert codename == value[1]
        assert extension == value[2]



from pathlib import Path

import pytest
import git
from termcolor import cprint

from time import time
from sitefab import __version__ as version
from sitefab import utils
from sitefab import files
from sitefab.SiteFab import SiteFab

# NOTE you don't need to import fixture explictly pytest do it for you. Avoid
# flake8 errors

TEST_ROOT_DIR = Path(__file__).parent
TEMPLATE_DATA_ROOT = TEST_ROOT_DIR / 'sitefab_template'

TEMPLATE_GIT_URL = 'https://github.com/ebursztein/sitefab-template'
PLUGINS_GIT_URL = 'https://github.com/ebursztein/sitefab-plugins'

# each excution have a new TEMPLATE_DATA_PATH to avoid cache
# and find git miss commit issues.
TEMPLATE_DATA_PATH = None
TEMPLATE_DATA_CONFIG_FILE_PATH = None
PLUGINS_DATA_PATH = None


@pytest.fixture()
def myobjdict():
    d = {'a': 'test', 'b': {"c": '2nd'}}
    return utils.dict_to_objdict(d)


@pytest.fixture(scope="session")
def sitefab():
    "pull needed data and create a valid instance of sitefab"

    fname = TEMPLATE_DATA_PATH / "config" / "sitefab.yaml"
    print('SiteFab version: %s' % version)
    return SiteFab(fname)


@pytest.fixture(scope="function")
def empty_post():
    "mock a post"
    post = utils.create_objdict()
    post.md = ""
    post.html = ""
    post.meta = utils.create_objdict()
    post.meta.statistics = utils.create_objdict()
    post.meta.toc = utils.create_objdict()
    post.elements = utils.create_objdict()
    return post


def pytest_configure(config):
    global TEMPLATE_DATA_PATH
    global TEMPLATE_DATA_CONFIG_FILE_PATH
    global PLUGINS_DATA_PATH

    TEMPLATE_DATA_PATH = TEMPLATE_DATA_ROOT / str(int(time()))
    TEMPLATE_DATA_CONFIG_FILE_PATH = TEMPLATE_DATA_PATH / 'config' / 'sitefab.yaml'  # noqa
    PLUGINS_DATA_PATH = TEMPLATE_DATA_PATH / 'plugins'

    cprint('[Cleanup]', 'magenta')
    cprint('|- deleting old site template: %s' % TEMPLATE_DATA_ROOT, 'cyan')
    try:
        files.clean_dir(TEMPLATE_DATA_ROOT)
    except:  # noqa
        cprint('Warning: part of directory not deleted - should be fine thus',
               'yellow')

    # recreate if correctly deleted
    if not TEMPLATE_DATA_ROOT.exists():
        TEMPLATE_DATA_ROOT.mkdir()

    cprint('[SiteFab Templates]', 'magenta')
    if TEMPLATE_DATA_PATH.exists():
        cprint('Pulling latest sitefab template', 'green')
        g = git.cmd.Git(TEMPLATE_DATA_PATH)
        g.pull()
    else:
        cprint('Cloning sitefab template', 'yellow')
        git.Repo().clone_from(TEMPLATE_GIT_URL, TEMPLATE_DATA_PATH)

    cprint('[SiteFab Plugins]', 'magenta')
    if PLUGINS_DATA_PATH.exists():
        cprint('Pulling latest sitefab plugins', 'green')
        g = git.cmd.Git(PLUGINS_DATA_PATH)
        g.pull()
    else:
        cprint('Cloning sitefab plugins', 'yellow')
        git.Repo().clone_from(PLUGINS_GIT_URL, PLUGINS_DATA_PATH)



import numpy as np
from perfcounters import PerfCounters
from tabulate import tabulate
from textacy import TextStats, keyterms, make_spacy_doc, preprocessing
from textacy.ke.yake import yake
from textacy.ke.scake import scake

from sitefab.utils import create_objdict

# FIXME: use the config
NUM_TERMS = 50
SPACY_MODEL = 'en_core_web_sm'  # 'en_core_web_lg'
# python -m spacy download en_core_web_sm
TERM_EXTRACTOR_ALGO = 'yake'  # yake, sgrank, textrank
NGRAMS = (1, 2, 3)  # default


def softmax(results, reverse=False):
    """Normalize results values via softmax.

    Args:
        results (array): term extraction results.
        reverse (bool, optional): [description]. Defaults to False.

    Returns:
        [type]: [description]
    """

    if len(np.asarray(results).shape) == 1:
        # !case when there are less than 3 words as the rank algo won't work
        fill_value = 1 / len(results)
        x = np.full(len(results), fill_value)
        results = [[t, 0] for t in results]
    else:
        x = np.array([i[1] for i in results])

    if reverse:
        x = 1 - x

    e_x = np.exp(x - np.max(x))
    scores = e_x / e_x.sum()

    normalized_terms = []
    for idx, s in enumerate(scores):
        normalized_terms.append([results[idx][0], float(s)])
    return normalized_terms


def extract_key_terms(doc, num_terms=50, ngrams=(1, 2, 3), algo='yake'):
    """Compute post most important terms

    This is particularly useful for the search and related posts

    Args:
        doc (Spacy.doc): Doc to extract terms from.
        num_terms (int, optional): How many terms to return. Defaults to 100.
        ngrams (int, optional): which size of ngrams to consider
        algo (str, optional): which algorithm to use to find key terms
    """
    if not len(doc):
        return []

    # special case
    if len(doc) < 3:
        return softmax(str(doc).split(' '))

    if algo == 'textrank':
        return softmax(keyterms.textrank(doc, n_keyterms=NUM_TERMS))
    elif algo == 'yake':
        return softmax(yake(doc, ngrams=ngrams, topn=NUM_TERMS),
                       reverse=True)
    elif algo == 'scake':
        return softmax(scake(doc, topn=NUM_TERMS))
    elif algo == 'sgrank':
        return softmax(keyterms.sgrank(doc, ngrams=ngrams,
                                       n_keyterms=NUM_TERMS))
    else:
        err = 'Unknown key term extraction method:%s' % algo
        raise Exception(err)


def text_cleanup(text):
    "cleanup our text"

    text = preprocessing.replace_emails(text, replace_with='')
    text = preprocessing.replace_urls(text, replace_with='')
    text = preprocessing.replace_hashtags(text, replace_with='')
    text = preprocessing.replace_phone_numbers(text, replace_with='')
    text = preprocessing.replace_numbers(text, replace_with='')

    text = preprocessing.remove_accents(text)
    text = preprocessing.remove_punctuation(text)

    text = preprocessing.normalize_quotation_marks(text)
    text = preprocessing.normalize_hyphenated_words(text)
    text = text.replace('\n', ' ').replace('\t', ' ')
    text = text.lower()

    text = preprocessing.normalize_whitespace(text)
    return text


def generate_clean_fields(post):
    "Generate a cleaned up version of the post and its metadata"
    clean_fields = create_objdict()

    # cleaned up fields
    clean_fields.title = ''
    if post.meta.title:
        clean_fields.title = text_cleanup(post.meta.title)

    clean_fields.abstract = ""
    if post.meta.abstract:
        clean_fields.abstract = text_cleanup(post.meta.abstract)

    clean_fields.authors = []
    if post.meta.authors:
        for author in post.meta.authors:
            clean_fields.authors.append(text_cleanup(author))

    # conference
    clean_fields.conference_name = []
    if post.meta.conference_name:
            clean_fields.conference_name = text_cleanup(
                post.meta.conference_name)

    clean_fields.conference_short_name = ""
    if post.meta.conference_short_name:
        clean_fields.conference_short_name = text_cleanup(
            post.meta.conference_short_name)

    # category, tags, etc
    clean_fields.category = ""
    if post.meta.category:
        clean_fields.category = text_cleanup(post.meta.category)

    clean_fields.tags = []
    if post.meta.tags:
        for tag in post.meta.tags:
            clean_fields.tags.append(text_cleanup(tag))

    # text
    clean_fields.text = ''
    if post.text:
        # !make sure to use post html and clean it to avoid markup keywords.
        clean_fields.text = text_cleanup(post.text)

    return clean_fields


def benchmark_term_extractor(doc, counters):
    "benchmark various term extractor algorithms"
    # TL;DR: yake is probably the best. Feel free to experiment
    # see https://github.com/LIAAD/yake
    # sgrank is really really slow
    results = []
    methods = ['textrank', 'yake', 'sgrank']  # 'cake',     results = []
    for method in methods:
        counters.start(method)
        results.append(extract_key_terms(doc, algo=method))
        counters.stop(method)

    counters.report()
    table = []
    for idx in range(20):
        row = []
        for aidx in range(len(methods)):
            row.append(results[aidx][idx])
        table.append(row)

    print(tabulate(table, headers=methods))
    return counters


def compute_stats(doc):
    ts = TextStats(doc)
    stats = create_objdict()
    stats.readability = create_objdict(ts.readability_stats)

    fields = {'n_sents': 'sentences',
              'n_words': 'words',
              'n_unique_words': 'unique_words',
              'n_chars': 'chars',
              'n_long_words': 'long_words',
              'n_syllables': 'syllables',
              'n_monosyllable_words': 'monsyllable_words',
              'n_polysyllable_words': 'pollysyllable_words'
              }

    stats.counts = create_objdict()
    for src, dst in fields.items():
        if src in ts.basic_counts:
            stats.counts[dst] = ts.basic_counts[src]
        else:
            stats.counts[dst] = 0

    return stats


def analyze_post(post, debug=False):
    "Perform NLP analysis"

    counters = PerfCounters()
    nlp = create_objdict()

    # clean fields
    counters.start('cleanup')
    clean_fields = generate_clean_fields(post)
    nlp.clean_fields = clean_fields
    counters.stop('cleanup')

    # creating spacy docs
    counters.start('make_spacy_docs')
    all_cleaned_content = ' '.join([clean_fields.title, clean_fields.category,
                                    " ".join(clean_fields.tags),
                                    clean_fields.abstract, clean_fields.text])

    # overall terms
    cleaned_doc = make_spacy_doc(all_cleaned_content, lang=SPACY_MODEL)

    # title terms
    title_doc = make_spacy_doc(clean_fields.title, lang=SPACY_MODEL)

    # for statistics
    text_doc = make_spacy_doc(post.text, lang=SPACY_MODEL)

    counters.stop('make_spacy_docs')

    # terms extraction
    counters.start('extract_key_terms')
    nlp.terms = extract_key_terms(cleaned_doc, num_terms=NUM_TERMS,
                                  algo=TERM_EXTRACTOR_ALGO, ngrams=NGRAMS)

    # !note we restrict ngram to one as we only want the lemmized top terms.
    nlp.title_terms = extract_key_terms(title_doc, num_terms=NUM_TERMS,
                                        algo=TERM_EXTRACTOR_ALGO, ngrams=1)

    counters.stop('extract_key_terms')

    # text stats
    counters.start('text_stats')
    nlp.stats = compute_stats(text_doc)
    counters.stop('text_stats')
    if debug:
        counters.report()
    return nlp



from .Plugins import Plugins  # noqa
from .CollectionProcessor import CollectionProcessor # noqa
from .PostProcessor import PostProcessor  # noqa
from .SitePreparsing import SitePreparsing  # noqa
from .SiteProcessor import SiteProcessor  # noqa
from .SiteRendering import SiteRendering  # noqa
from .TemplateFilter import TemplateFilter  # noqa



import os
from tqdm import tqdm

from . import utils, files


class PostCollections():
    "Handle posts collections"

    def __init__(self, site, template=None, output_path=None, web_path=None,
                 min_posts=0):
        """Init the collections

            Args:
                site (SiteFab): current instance of SiteFab
                template (Template): the jinja2 template used to render the collections. If none can't be rendered (optional)
                path (str): directory where to render the collections. If none can't be rendered (optional)
                min_posts (int): minimal number of posts to render the category (optional)
        """

        self.site = site
        self.collections = {}
        self.template = template
        self.output_path = output_path
        self.web_path = web_path
        self.min_posts = min_posts

    def add(self, name, post):
        """Create a Collection

        Args:
            name (str): name of the Collection.
            post (Post): post to add to the collection
        Return:
            None
        """

        if name not in self.collections:
            collection = utils.dict_to_objdict()
            collection.posts = []

            collection.meta = utils.dict_to_objdict()
            collection.meta.name = name
            collection.meta.num_posts = 0
            if self.web_path:
                url = "/%s%s" % (self.web_path, name)
                url = url.replace(" ", "-").lower()
                collection.meta.url  = url
            self.collections[name] = collection
        self.collections[name].meta.num_posts += 1
        self.collections[name].posts.append(post)

    def render(self):
        "Render collections pages."

        for collection in tqdm(self.get_as_list(), unit=' collections', miniters=1, desc="Collections"):
                collection.meta.slug = collection.meta.name.replace(" ", "-").lower()
                if collection.meta.num_posts >= self.min_posts:
                    filename = "%s.html" % (collection.meta.slug)
                    rv = self.template.render(posts=collection.posts, meta=collection.meta, plugin_data=self.site.plugin_data, config=self.site.config)
                    new_path = os.path.join(self.output_path, collection.meta.slug)
                    files.write_file(new_path, "index.html", rv)

    def get_as_list(self):
        """Returns the collections as lists
        Return:
            list: collections as list
        """
        return self.collections.values()

    def get_as_dict(self):
        """ return the collections as dict
        Return:
            dict: collections as dict
        """

        return self.collections

    def get_num_collections(self):
        """ return the number of collections
        Return:
            int: number of collections
        """
        return len(self.collections)



class SitePreparsing():
    """Site wide plugins that execute before the parsing start.
    Plugin are called only once."""

    def process(self, unused, site, config):
        """ Process the content of the site once before parsing

        Args:
            unused (None): not used.
            site (SiteFab): instanciated site object.
            config (objectdict): plugin configuration.
        """



from sitefab import utils


def test_creation(myobjdict):
    d = {'a': 'test', 'b': {"c": '2nd'}}
    p = utils.dict_to_objdict(d)
    assert d['a'] == p.a
    assert d['b']['c'] == p.b.c


def test_insertion(myobjdict):
    myobjdict.test = 'yes'
    assert myobjdict.test == 'yes'
    assert myobjdict['test'] == 'yes'
    myobjdict['test2'] = 'oui'
    assert myobjdict.test2 == 'oui'


def test_deletion(myobjdict):
    myobjdict.test = 'yes'
    del myobjdict['test']
    assert 'yes' not in myobjdict



from .utils import get_linter_errors_list
from sitefab.utils import objdict


def test_no_meta(sitefab, empty_post):
    empty_post.meta = None
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert isinstance(error_list, list)


def test_meta_no_toc(sitefab, empty_post):
    empty_post.meta = objdict()
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert isinstance(error_list, list)


def test_e300_triggered(sitefab, empty_post):
    empty_post.meta.toc = [["headline", 1, 0]]
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E300" in error_list


def test_e300_not_triggered(sitefab, empty_post):
    empty_post.meta.toc = [["headline", 2, 0]]
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E300" not in error_list


def test_e301_triggered(sitefab, empty_post):
    empty_post.meta.toc = [["headline", 2, 0]]
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E301" in error_list


def test_e301_not_triggered(sitefab, empty_post):
    empty_post.meta.toc = [["headline", 2, 0], ["headline 2", 2, 1]]
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E301" not in error_list



from pathlib import Path
from jinja2 import Template

from sitefab import utils
from sitefab import files
from . import frontmatter, images, structure


class Linter:

    def __init__(self, config):
        current_dir = Path(__file__).parent
        test_file = current_dir / 'tests.yaml'
        self.test_info = files.load_config(test_file)
        if not self.test_info:
            utils.error("Can't load linter tests")

        self.config = config
        self.results = {}
        template_content = files.read_file(self.config.report_template_file)
        self.jinja2_template = Template(str(template_content))

    def render_report(self):
        "Create a linting report for all posts"
        report = self.jinja2_template.render(results=self.results)
        files.write_file(self.config.output_dir, "linter.html", report)

    def num_post_with_errors(self):
        "Return number of posts that have errors"
        cnt = 0
        for p in self.results.values():
            cnt += p.has_errors
        return cnt

    def num_post_with_warnings(self):
        "Return the numbers of posts that have warnings but no errors"
        cnt = 0
        for p in self.results.values():
            if not p.has_errors:
                cnt += p.has_warnings
        return cnt

    def num_post_ok(self):
        "num post with no wrror or warnings"
        cnt = 0

        for p in self.results.values():
            if not p.has_errors and not p.has_warnings:
                cnt += 1
        return cnt

    def lint(self, post, rendered_post, site):
        """ Load yaml configuration

        Args:
            post (Post): the post to analyze
            rendered_post (str): the html version of the post
            site (Sitefab): the site object mainly used to get access
            to plugin data
        Return:
            dict: linting results
        """
        results = utils.create_objdict()
        results.has_errors = 0
        results.has_warnings = 0

        # frontmatter
        results.info = frontmatter.lint(post, self.test_info, self.config)

        # images
        if 'image_info' in site.plugin_data:
            image_info = site.plugin_data['image_info']
        else:
            image_info = None

        img_results = images.lint(post, self.test_info, self.config,
                                  image_info)
        results.info.extend(img_results)
        stucture_results = structure.lint(post, self.test_info, self.config)
        results.info.extend(stucture_results)

        for d in results.info:
            if d[0][0] == "E":
                results.has_errors += 1
            if d[0][1] == "W":
                results.has_warnings += 1

        if results.has_errors or results.has_warnings:
            self.results[post.filename] = results

        return results



from sitefab.files import get_files_list


def test_parser_template_path(sitefab):
    "making sure we have the right templates in the parser"
    assert 'a' in sitefab.config.parser.templates
    assert 'img' in sitefab.config.parser.templates
    assert 'youtube' in sitefab.config.parser.templates


def test_parser_templates_loaded(sitefab):
    print(sitefab.config.parser.templates_path)
    assert len(get_files_list(sitefab.config.parser.templates_path, "*.html"))



class TemplateFilter():
    "Plugins that define jinja2 filters to be used in templates"

    @staticmethod
    def myfilter(filter_input, filter_arg):
        """Act as a jinja2 template

        Args:
            filter_input (str): the input passed to the filter
            filter_arg(str): optional arg to the filter
        Return:
            str: modified input
        """



from pathlib import Path
from sitefab.files import read_file, write_file, clean_dir, get_files_list


def test_basic_file_flow(tmp_path):
    """testing all at once as it is hard to do in smaller chunk
       write -> read -> list -> clean -> list again
    """
    fname = 'a.txt'
    content = 'this is a test'
    path = Path(tmp_path) / fname

    # write
    write_file(tmp_path, fname, content)
    assert path.exists()

    # read
    read_back = read_file(path)
    assert read_back == content

    # list
    files_list = get_files_list(tmp_path, '*.txt')
    assert path in files_list

    # clean
    clean_dir(tmp_path)
    assert not path.exists()

    # filelist again
    files_list = get_files_list(path, '*.txt')
    assert not files_list  # must be empty


def test_unicode_support(tmp_path):

    fname = '😁.txt'
    content = 'this is a test 😁'
    path = Path(tmp_path) / fname

    # write
    write_file(tmp_path, fname, content)
    assert path.exists()

    # read
    read_back = read_file(path)
    assert read_back == content

    # cleanup
    clean_dir(tmp_path)


def test_writing_subdirectory(tmp_path):
    "make sure sub directory is created"

    fname = 'myfile.md'
    content = 'this is a test 😁'
    subdir = 'tobecreated'
    subdir_path = Path(tmp_path) / subdir
    path = subdir_path / fname

    # write
    write_file(subdir_path, fname, content)
    assert path.exists()
    assert subdir_path.is_dir()

    # read
    read_back = read_file(path)
    assert read_back == content

    # cleanup
    clean_dir(tmp_path)
    assert not path.exists()
    assert not subdir_path.exists()


def test_non_existing_read_file():
    assert read_file('asjkajlkajlksal32312ewdas3') == ""


def test_filelist_filter(tmp_path):
    "making sure we get all the files which have a givem suffix"
    fnames = ['😁.md', 'a.m', 'a.txt', 'b.md', 'a.md.md']
    content = 'this is a test'

    clean_dir(tmp_path)
    for fname in fnames:
        write_file(tmp_path, fname, content)

    files_list = get_files_list(tmp_path, '*.md')
    assert len(files_list) == 3

    files_list = get_files_list(tmp_path, '*.txt')
    assert len(files_list) == 1

    clean_dir(tmp_path)


def test_filelist_filter_empty(tmp_path):
    "ensure no matches return an empty array"
    clean_dir(tmp_path)
    assert get_files_list(tmp_path) == []
    assert get_files_list('thisisnotthedirectoryyoulookfor') == []


def test_filelist_recursive(tmp_path):
    clean_dir(tmp_path)
    sub_path = tmp_path / 'newdir'
    fname = 'test.md'

    txt = 'hello'
    write_file(sub_path, 'test.md', txt)

    # ensure we have a sub dir and the file in it
    assert sub_path.exists()
    assert sub_path.is_dir()
    read_text = read_file(sub_path / fname)
    assert read_text == txt

    # test we get a recursive listing
    assert len(get_files_list(tmp_path)) == 1

    # test that non recursive returns nothing
    assert get_files_list(tmp_path, recursive=False)  == []



"Post parser"
import jinja2
import mistune
from pygments.formatters import html

from sitefab import files, utils
from sitefab.parser import frontmatter
from sitefab.parser.html2text import html2text
from sitefab.parser.markdown import HTMLRenderer


class Parser():

    def __init__(self, config, site):
        """ Initialize a new parser for a given type of parsing

        Args:
            config (obj_dict): the parser config. It is explicitly defined to
            allows  different conf
            site (dictobj): sitefab instantiated object
        Return:
            None

        note: one parser is instanciated by thread. potentially batching more
        than one post per thread might help with performance.
        """

        # verify that the config exist
        self.config = config
        self.site = site

        # NOTE: Might seems weird to do this but templates is what allows
        # to have different rendering for each plugins. So we keep it explict
        # in the code as this got messed up countless time :(
        self.templates = self.config.templates

        # templates are not compiled yet, they will be when parsing will be
        # called the first time
        self.jinja2 = None

        # markdown parser
        self.renderer = HTMLRenderer()
        self.md_parser = mistune.Markdown(renderer=self.renderer)

        # code higlighterr
        if self.config.code_display_line_num:
            linenos = 'table'
        else:
            linenos = False

        self.code_formatter = html.HtmlFormatter(
            style=self.config.code_highlighting_theme,
            nobackground=False, linenos=linenos)

    @staticmethod
    def make_config(config):
        """ Initialize a parser config with all the needed variables

        Args:
            config (obj_dict): the uninitialized configuration with basic
            variables
        Returns:
            obj_dict: The initialized configuration
        """

        if not config:
            utils.detailed_error("Parser", 'make_config',
                                 'supplied config is empty')

        if 'template_dir' not in config:
            utils.detailed_error("Parser", 'make_config',
                                 'template_dir not found')

        config.templates = {}
        for fname in files.get_files_list(config.templates_path, "*.html"):
            template = files.read_file(fname)
            config.templates[fname.stem] = template
        return config

    def list_templates(self):
        "Return the list of available templates"
        return self.templates.keys()

    def parse(self, md_file):
        """ Parse a md file into a post object
        """

        # compile the templates when we parse the first post. This is needed
        # to ensure that plugins get a chance to modify the templates before
        # we compile them.
        if not self.jinja2:
            self.jinja2 = jinja2.Environment(loader=jinja2.DictLoader(
                                             self.templates))

        parsed_post = utils.dict_to_objdict()

        # parsing frontmatter and getting the md
        parsed_post.meta, parsed_post.md = frontmatter.parse(md_file)

        # parsing markdown and extractring info
        # NOTE: this must called before every parsing
        self.renderer.init(self.jinja2, self.code_formatter, self.site,
                           parsed_post.meta)

        parsed_post.html = self.md_parser.parse(parsed_post.md)
        parsed_post.text = html2text(parsed_post.html)  # used by NLP
        parsed_post.meta.statistics = self.renderer.get_stats()
        parsed_post.meta.toc = self.renderer.get_json_toc()
        parsed_post.elements = self.renderer.get_info()
        return parsed_post



from sitefab import utils
from sitefab.utils.objdict import objdict


def test_dict_to_objdict():
    dic = {
        "str": "str",
        "int": 1,
        "array": [1, 2, 3]
    }
    objdict = utils.dict_to_objdict(dic)
    assert objdict.str == "str"
    assert objdict.int == 1
    assert isinstance(objdict.array, type([]))


def test_create_objdict():
    od = utils.create_objdict()
    od2 = objdict()
    assert type(od) == type(od2)


def test_objdict_to_dict():
    od = utils.create_objdict()
    od.str = "str"
    od.int = 1
    od.array = [1, 2, 3]
    od.arrayofarray = [[1, 2], [3, 4]]
    d = utils.objdict_to_dict(od)
    assert d['str'] == "str"
    assert d['int'] == 1
    assert d['array'][2] == 3
    assert d['arrayofarray'][0][0] == 1   # testing nested array
    assert d['arrayofarray'][1][1] == 4   # testing nested array




from sitefab import files


def generate_plugins_readme(site, output_fname):
    "Generate plugins readme"

    plugin_list_md = "# SiteFab plugins\n"
    plugin_list_md += "List of available plugins\n\n"
    plugin_list_md += "|Name | Description | dependencies|\n"
    plugin_list_md += "|-----|:------------|:------------|\n"

    plugins = site.plugins.get_plugins()
    for plugin in plugins:
        name = plugin.name
        description = plugin.description
        dependencies = site.plugins.get_plugin_dependencies(plugin)
        if not len(dependencies):
            dependencies = ""
        else:
            dependencies = ", ".join(dependencies)
        doc_path = site.plugins.get_plugin_documentation_filename(plugin)
        doc_path = str(doc_path)
        if 'plugins' in doc_path:
            doc_path = doc_path.replace('\\', '/').split('plugins')[1]
            name = "[%s](tree/master%s)" % (name, doc_path)
        plugin_list_md += "| %s | %s | %s |\n" % (name, description,
                                                  dependencies)

    files.write_file('', output_fname, plugin_list_md)



import pytest
from sitefab.cmdline.cmdline import generate, version
from sitefab.files import get_files_list
from ..conftest import TEMPLATE_DATA_PATH


@pytest.mark.dependency(name="compile")
def test_compile_and_check_generated_content(sitefab):
    generate(sitefab.config_filename, version)


@pytest.mark.dependency(name="log", depends=["compile"])
def test_logs_presence():
    log_dir = TEMPLATE_DATA_PATH / 'logs'
    fnames = ['index.html', 'linter.html', 'stats.html']
    for fname in fnames:
        path = log_dir / fname
        assert path.exists()
        assert path.is_file()


@pytest.mark.dependency(name="content", depends=["compile"])
def test_content_presence():
    content_dir = TEMPLATE_DATA_PATH / 'generated'

    # root files
    fnames = ['index.html', 'rss.xml', 'sitemap.xml']
    for fname in fnames:
        path = content_dir / 'index.html'
        assert path.exists()
        assert path.is_file()

    # files in directories
    dirnames = ['blog', 'code', 'publications']
    for dirname in dirnames:
        path = content_dir / dirname / 'index.html'
        assert path.exists()
        assert path.is_file()


# coding: utf-8
"""
SiteFab: content > sitefab > static site
"""
import sys
import getopt
from collections import defaultdict

from sitefab import __version__ as version
from sitefab.SiteFab import SiteFab
from termcolor import colored, cprint
from sitefab.utils import print_color_list, section, print_header
from sitefab.docs.plugins import generate_plugins_readme


def print_plugins_list(site, only_enable=True):
    "Output the list of plugins"
    info = defaultdict(lambda: defaultdict(list))
    categories = ['Collection', 'Post', 'Site']
    phases = ['Preparsing', 'Parsing', 'Processor', 'Rendering']

    for pl in site.plugins.get_plugins_info():
        cat = pl[0]
        name = pl[1]
        status = pl[3]
        desc = pl[2]
        version = pl[5]

        if only_enable:
            if not status:
                continue
            status = ""
        else:
            if status:
                status = colored("[enable]  ", 'green')
            else:
                status = colored("[disable] ", 'red')

        s = "%-20s (%s):\t %s" % (name, version, desc)

        for category in categories:
            if category in cat:
                phase = cat.replace(category, "")
                info[category][phase].append([status, s])

    for category, data in info.items():
        print(colored("[%s]" % category, 'yellow'))
        for phase in phases:
            if len(data[phase]):
                print(colored("  %s" % phase, 'magenta'))
                count = 0
                for plugin in data[phase]:
                    if count % 2:
                        plugin[1] = colored(plugin[1], 'blue')
                    else:
                        plugin[1] = colored(plugin[1], 'cyan')
                    print("    |-%s%s" % (plugin[0], plugin[1]))
                    count += 1
        print("\n")


def generate(config, version):
    "generate command main function"
    section("Init")
    # initializing site
    site = SiteFab(config, version)
    cprint("Directories", 'magenta')
    dirs = []
    dirs.append("Ouput:\t%s" % site.get_output_dir())
    dirs.append("Logs:\t\t%s" % site.get_logs_dir())
    dirs.append("Content:\t%s" % site.get_content_dir())
    dirs.append("Template:\t%s" % site.get_template_dir())
    dirs.append("Plugins:\t%s" % site.get_plugins_dirs())
    dirs.append("Cache:\t%s" % site.get_cache_dir())
    print_color_list(dirs)
    print("\n")
    cprint("Active plugins", 'magenta')
    print_plugins_list(site)

    section("Pre processing")
    site.preprocessing()
    # parsing content into post objects
    section("Parsing")
    site.parse()

    # processing posts and collections to add extra info
    section("Processing")
    site.process()

    section("Rendering")
    site.render()
    # Generate auxiliary files (sitemap, facebook_instant etc)
    # FIXME Cleanup the output
    section("Summary")
    site.finale()


def print_help():
    "Display help and exist"

    cprint("Usage: sitefab -c <config_file> command", 'yellow')

    # end user
    cmds = [
        "generate: generate the site",
        "plugins: list available plugins",
        ]

    cprint("Available Commands", 'magenta')
    print_color_list(cmds, prefix="\t")

    # dev
    #  dev_cmds = [
    #     "gen_plugins_readme: generate plugins README.md from plugins infos",
    # ]
    # cprint("Developper command", 'blue')
    # print_color_list(dev_cmds, prefix="\t")

    sys.exit(2)


def main():
    config = None
    short_options = "c:h:o:"
    long_options = ["config=", "help", "output_file="]

    # pretty banner
    print_header(version)

    # parsing
    try:
        options, args = getopt.getopt(sys.argv[1:], short_options,
                                      long_options)
    except getopt.GetoptError:
        print_help()

    # options
    for opt, arg in options:
        if opt in ('-c', '--config'):
            config = arg
        elif opt in ('-h', '--help'):
            print_help()
        elif opt in ('-o', '--output_file'):
            # used for documentation generation
            output_fname = arg
    # arguments
    if len(args):
        cmd = args[0]

        if not config:
            cprint('[ERROR] Please provide a config file via the -c option',
                   'red')
            print_help()

        if cmd == "generate":
            generate(config, version)

        elif cmd == "plugins":
            site = SiteFab(config)
            cprint("Plugins status", 'magenta')
            print_plugins_list(site, only_enable=False)

        # doc command
        elif cmd == "gen_plugins_readme":
            # this function rebuild the plugin readme
            site = SiteFab(config, version)
            generate_plugins_readme(site, output_fname)

        else:
            print_help()
    else:
        print_help()


if __name__ == '__main__':
    main()



from sitefab.parser.html2text import html2text


def test_none():
    assert html2text(None) == ''


def test_empty():
    assert html2text('') == ''


def test_basic():
    assert html2text('<span>test</span>') == 'test'


def test_pre():
    snip = '''
        <pre>alert(1)</pre>
        <span>test</span>
    '''
    assert html2text(snip) == 'test'


def test_code():
    snip = '''
        <code>alert(1)</code>
        <span>test</span>
    '''
    assert html2text(snip) == 'test'


def test_upercase_code():
    snip = '''
        <code>alert(1)</CoDe>
        <span>test</span>
    '''
    assert html2text(snip) == 'test'


def test_dangling_tags():
    snip = '''
        <table><td>
            <span>test</span>
    '''
    assert html2text(snip) == 'test'


def test_broken_html():
    snip = '''
        </td>test</span>
    '''
    assert html2text(snip) == 'test'



# coding: utf-8
import re

# syntax coloring
from pygments import highlight
from pygments.lexers import get_lexer_by_name, guess_lexer


from mistune import Renderer, escape
from sitefab import utils

youtube_matcher = re.compile("v=([^&]+)")


class HTMLRendererMixin(object):
    """Customized HTML renderer"""
    def link(self, link, title, content):
        embed = False
        src = link

        # Youtube
        if (("https://youtu.be/" in link or "https://www.youtube.com/" in link)
                and ("&no_embed=1" not in link)):
            embed = True
            if "embed" in link:  # Already correct link
                src = link
                template = self.jinja2.get_template('youtube')
            else:
                # need to canonalize youtube url
                if "https://youtu.be/" in link:
                    src = "https://www.youtube.com/embed/"
                    src += link.replace("https://youtu.be/", "")
                else:
                    d = youtube_matcher.search(link)
                    if d:
                        vid = d.group(1)
                        src = "https://www.youtube.com/embed/" + vid
                    else:
                        print("error can't detect videoid for link: %s" % link)
                        print(self.meta.title)
                template = self.jinja2.get_template('youtube')
                self.info.videos.append(src)

        # Normal links or not embedded youtube videos
        else:
            src = link.replace("&no_embed=1", "")
            template = self.jinja2.get_template('a')
            self.info.links.append(src)

        rv = template.render(href=src, text=content, title=title, embed=embed,
                             site=self.site, meta=self.meta)
        # rv = rv.encode('utf-8')
        return rv

    def image(self, src, title, alt_text):

        self.info.images.append(src)
        template = self.jinja2.get_template('img')
        rv = template.render(src=src, alt=alt_text, title=title,
                             plugin_data=self.plugin_data, site=self.site,
                             meta=self.meta)
        # rv = rv.encode('utf-8')
        return rv

    def header(self, text, level, raw=None):

        template = self.jinja2.get_template('h')
        rv = template.render(level=level, text=text, id=self.toc_count,
                             site=self.site, meta=self.meta)
        # rv = rv.encode('utf-8')

        self.toc_tree.append((self.toc_count, text, level, raw))
        self.toc_count += 1
        return rv

    def block_quote(self, text):
        "Block quote highlighter"
        template = self.jinja2.get_template('blockquote')
        rv = template.render(text=text, site=self.site, meta=self.meta)
        # rv = rv.encode('utf-8')
        return rv

    def block_code(self, code, lang):
        "Block code highlighter and formater"
        try:
            if not lang:
                lexer = guess_lexer(code, stripall=True)
            else:
                lexer = get_lexer_by_name(lang, stripall=True)
            code = highlight(code, lexer, self.code_formatter)
        except:  # noqa
            code = escape(code)
            lang = None

        self.info.code.append(code)

        template = self.jinja2.get_template('code')
        rv = template.render(code=code, lang=lang, site=self.site,
                             meta=self.meta)
        # rv = rv.encode('utf-8')
        return rv

    def init(self, jinja2, code_formatter, site, meta):
        """Init function called before each parsing.

        Args:
            jinja2 (dict): jinja2 templates used for rendering
            code_formatter (dict): code syntax highlight configuration
            site (obj_dict): the full site context (SiteFab object)
            meta (obj_dict): the meta associated with the post
        Return:
            None
        Note:
            Used to ensure all the needed variables are reset between parsing
            executions.
        """
        # reset toc
        self.toc_tree = []
        self.toc_count = 0
        self.jinja2 = jinja2
        self.code_formatter = code_formatter
        self.plugin_data = site.plugin_data
        self.site = site
        self.meta = meta

        # Various information collected during the parsing
        self.info = utils.dict_to_objdict({
            "links": [],
            "images": [],
            "videos": [],
            "code": []
        })

        self.stats = utils.dict_to_objdict({
            "num_links": 0,
            "num_images": 0,
            "num_videos": 0,
            "num_code": 0,
        })

    def get_info(self):
        return self.info

    def get_stats(self):
        stats = self.stats
        stats['num_links'] = len(self.info['links'])
        stats['num_videos'] = len(self.info['videos'])
        stats['num_images'] = len(self.info['images'])
        return stats

    # [TOC code] #
    def get_json_toc(self):
        """Render the TOC in JSON.

        Returns:
            list: TOC as a list ready to be json.dumps().
        """
        lst = []
        for toc in self.toc_tree:
            index, text, l, raw = toc
            lst.append((text, l, index))
        return lst

    def get_html_toc(self, level=3):
        """Render TOC as HTML table.

        Args:
            level (int, optional): Max TOC level. Defaults to 3.

        Returns:
            str: HTML table containing the TOC.
        """
        return ''.join(self._iter_html_toc(level))

    def _iter_html_toc(self, level):
        first_level = None
        last_level = None

        yield '<ul id="table-of-content">\n'

        for toc in self.toc_tree:
            index, text, l, raw = toc

            if l > level:
                # ignore this level
                continue

            if first_level is None:
                # based on first level
                first_level = l
                last_level = l
                yield '<li><a href="#toc-%d">%s</a>' % (index, text)
            elif last_level == l:
                yield '</li>\n<li><a href="#toc-%d">%s</a>' % (index, text)
            elif last_level == l - 1:
                last_level = l
                yield '<ul>\n<li><a href="#toc-%d">%s</a>' % (index, text)
            elif last_level > l:
                # close indention
                yield '</li>'
                while last_level > l:
                    yield '</ul>\n</li>\n'
                    last_level -= 1
                yield '<li><a href="#toc-%d">%s</a>' % (index, text)

        # close tags
        yield '</li>\n'
        while last_level > first_level:
            yield '</ul>\n</li>\n'
            last_level -= 1

        yield '</ul>\n'


class HTMLRenderer(HTMLRendererMixin, Renderer):
    pass



"files manipulation utilities"
import codecs
import shutil
from pathlib import Path
from time import sleep

import yaml

from . import utils as utils


def get_code_path():
    """Get SiteFab base directory path

    Note:
        code path is the dir where we have: sitefab, plugins... it's not
        the package root
    """
    path = Path(__file__)
    return(path.parents[1])


def load_config(config_file):
    """ Load yaml configuration

    Args:
        config_file (str): The path to the config file to load.

    Returns:
        objdict: configuration file parsed as dict or empty if parsing fail
    """

    # open file
    config_file = Path(config_file)
    if not config_file.exists:
        st = "%s does not exist" % config_file
        utils.warning(st)
        return None
    stream = open(config_file, 'r')

    # parse  yaml
    try:
        config_yaml = yaml.load(stream, Loader=yaml.SafeLoader)
    except yaml.YAMLError as exc:
        print(exc)
        return None

    # Making config a nice object
    config = utils.dict_to_objdict(config_yaml)
    return config


def read_file(filename):
    """Read a file and return its content. Use cached version unless specified.

    Args:
        filename (str): the filename to read from

    Returns:
        str: the content of the file
    """

    filename = Path(filename)
    if filename.is_file():
        f = codecs.open(filename, "r", "utf-8-sig")
        content = f.read()
        return content
    else:
        utils.warning("file:%s don't exist" % filename)
        return ""


def write_file(target_path, filename, content, binary=False):
    """ Write a file at a given path. Create directory if necessary

    Args:
        target_path (str): path where to write.
        filename (str): filename to write to.
        content (str or bytes): Content to write
        binary (bool, optional): Write as binary?. Defaults to False.

    Returns:
        None: no returns
    """
    target_path = Path(target_path)

    if not target_path.exists():
        target_path.mkdir(parents=True)

    file_path = target_path / filename
    if not binary:
        f = codecs.open(file_path, "w", "utf-8-sig")
        f.write(content)
        f.close()
    else:
        f = open(file_path, "wb")
        f.write(content)
        f.close()


def get_files_list(content_dir, extensions="*.md", recursive=True):
    """ Return the list of files in a directory and its sub directories that
    match a set of extensions.

    Args:
        content_dir (str): file path where content is located.
        extensions(str or list): single extension like "*.md" or array of
        extensions ["*.jpg", "*.png"]
        rescursive (bool): perform recursive listing

    Returns:
        list: list of content filename.

    Notes:
        matching is done the Usenix file matching way much like ls
    """

    # normalize the input
    if type(extensions) == str:
        extensions = [extensions]

    matches = []
    content_dir = Path(content_dir)
    for extension in extensions:
        if "*." not in extension:
            utils.warning("extension: %s won't match missing '*.'" % extension)
            continue
        if recursive:
            flist = list(content_dir.rglob(extension))
        else:
            flist = list(content_dir.glob(extension))
        matches.extend(flist)

    return matches


def clean_dir(directory):
    """Clean by deleting and recreating the directory

    Args:
        directory (str): Directory to clean.
    """
    directory = Path(directory)
    if directory.exists:
        shutil.rmtree(directory, ignore_errors=True)
        sleep(0.2)  # need time for the filesystem to register deletion
    if directory.exists():
        utils.error("%s not deleted." % directory)
    directory.mkdir()



from pathlib import Path


def test_config_paths(sitefab):
    # path
    partial_correct_path = Path('config/templates/logs')
    assert str(partial_correct_path) in str(sitefab.logger.config.template_dir)

    # templates
    assert sitefab.logger.config.log_template == 'log.html'
    assert sitefab.logger.config.log_index_template == "log_index.html"
    assert sitefab.logger.config.stats_template == "stats.html"


def test_directory_exist(sitefab):
    output_dir = Path(sitefab.logger.config.output_dir)
    assert output_dir.exists()
    assert output_dir.is_dir()



import re
from html.parser import HTMLParser
from textacy import preprocessing


class MLStripper(HTMLParser):
    def __init__(self):
        super().__init__()
        self.reset()
        self.fed = []

    def handle_data(self, d):
        self.fed.append(d)

    def get_data(self):
        return ''.join(self.fed)


def html2text(html):
    """HTML to text converter

    Args:
        html (str): html

    Returns:
        str: html page content in plaintext
    """
    if not html:
        return ''

    # remove code snippets
    html = re.sub(r'<pre>.*?</pre>', ' ', html,
                  flags=re.DOTALL | re.MULTILINE | re.IGNORECASE)
    html = re.sub(r'<code>.*?</code>', ' ', html,
                  flags=re.DOTALL | re.MULTILINE | re.IGNORECASE)

    # strip the rest
    s = MLStripper()
    s.feed(html)
    text = s.get_data()
    text = preprocessing.normalize_whitespace(text)
    return text



import sys
import xxhash

from .objdict import objdict
from termcolor import cprint


# hashing
def hexdigest(w):
    # xxhash is 3x/5x faster than md5
    return xxhash.xxh64(w).hexdigest()


# [colored output] #
def warning(txt):
    cprint("\n[Warning] %s\n" % txt, 'yellow')


def error(error_msg):
    cprint("\n[Error] %s" % error_msg, 'red')
    raise Exception(error_msg)


def detailed_error(module, function, error_msg):
    """ Display detail error and exit

    Args:
        module (str): the module where the error occured
        function (str): the function that triggred the errror message
        error_msg (str): the error message to display

    Return:
        None
    """
    msg = "\n[Error] %s:%s: %s\n" % (module, function, error_msg)
    cprint(msg)
    sys.exit(-42)


def section(val):
    cprint("\n[%s]" % val, 'yellow')


def print_color_list(lst, prefix='|-'):
    count = 0
    for i in lst:
        if count % 2:
            color = 'blue'
        else:
            color = 'cyan'
        st = "%s%s" % (prefix, i)
        cprint(st, color)
        count += 1


# [Object dict]
def create_objdict(dictionnary=None):
    "Create a an empty objdict object"
    if dictionnary:
        return dict_to_objdict(dictionnary)
    else:
        return objdict()


def dict_to_objdict(dictionnary=None):
    """ Convert a dict struct into a objdict structure

    Args:
        dictionnary (dict): the dictionnary to convert

    Returns:
        objdict: the dictionnary converted in objdict
    """
    o = objdict()
    if dictionnary:
        for k, v in dictionnary.items():
            if type(v) == dict:
                o[k] = dict_to_objdict(v)
            else:
                o[k] = v
    return o


def objdict_to_dict(objdict):
    """ Convert an objdict structure into a dict structure

    Args:
        obj (objdict): the objdict to convert

    Returns:
        dict: the objdict as standard dictionnary
    """
    d = {}
    if objdict:
        for k, v in objdict.items():
            if type(v) == dict:
                d[k] = objdict_to_dict(v)
            else:
                d[k] = v
    return d


def print_header(version):
        cprint('''
 ad88888ba   88                      88888888888          88
d8"     "8b  ""    ,d                88                   88
Y8,                88                88                   88
`Y8aaaaa,    88  MM88MMM  ,adPPYba,  88aaaaa  ,adPPYYba,  88,dPPYba,
  `"""""8b,  88    88    a8P_____88  88"""""  ""     `Y8  88P'    "8a
        `8b  88    88    8PP"""""""  88       ,adPPPPP88  88       d8
Y8a     a8P  88    88,   "8b,   ,aa  88       88,    ,88  88b,   ,a8"
 "Y88888P"   88    "Y888  `"Ybbd8"'  88       `"8bbdP"Y8  8Y"Ybbd8"'   v: %s
        ''' % version, "blue")
        cprint('''              -= https://github.com/ebursztein/sitefab =-
        ''', "cyan")



class SiteProcessor():
    "Plugins that process the whole site once during processing stage"

    def process(self, unused, site, config):
        """ Process the content of the site once during processing stage.

        Args:
            unused (None): not used.
            site (SiteFab): instanciated site object.
            config (objectdict): plugin configuration.
        """



"Image function common to many plugins"
from io import BytesIO
from .utils import hexdigest


def image_hash(raw_image):
    """Compute imgae hash

    Args:
        raw_image (bytes): image raw bytes.

    Returns:
        str: image hash
    """
    return hexdigest(raw_image)


def convert_image(img, extension_codename, compression_level=9,
                  jpeg_quality=85, webp_quality=85, return_as_bytesio=True):
    """Generate an image in the requested format

    Args:
        img (PIL.Image): The image to save.

        extension_codename (PIL extension name): normalize extension name.
        See: get_img_extension_alternative_naming to generate those

        compression_level (int, optional): Webp, png compression level.
        Defaults to 9.

        jpeg_quality (int, optional): Jpeg quality level. Defaults to 85.

        webp_quality (int, optional): Webp quality level. Defaults to 85.

        save (bool, optional): save image in a byteIO, if false return the
        image it self.
        Defauls to True.

    Returns:
        BytesIO or Image: the converted image as byteIO if return_as_bytesio
        is true or as Image otherwise.
    """

    img_io = BytesIO()

    # encoding
    if extension_codename == 'PNG':
        if return_as_bytesio:
            img.save(img_io, extension_codename, optimize=True,
                     compress_level=compression_level)

    elif extension_codename == 'WEBP':
        if img.mode == "P":
            img = img.convert('RGBA')
        if img.mode == "L":
            img = img.convert('RGB')
        img.save(img_io, extension_codename, optimize=True,
                 compress_level=compression_level, quality=webp_quality)

    elif extension_codename == "JPEG":
        if img.mode != "RGB":
            img = img.convert('RGB')
        img.save(img_io, extension_codename, optimize=True,
                 quality=jpeg_quality)

    elif extension_codename == "GIF":
        img.save(img_io, extension_codename, optimize=True)

    if return_as_bytesio:
        return img_io
    else:
        return img


def read_image_bytes(path):
    """Read raw image bytes from disk

    we need this because use a lot of bytesIO for manipulation
    and getting the bytes via PIL image API is 10x slower.

    Args:
        path (Path): image path

    Returns:
        bytes: images bytes
    """
    f = open(path, 'rb')
    raw_image = f.read()
    f.close()
    return raw_image


def save_image(img_io, path):
    """Save the image

    Args:
        img_io (BytesIO): The image byteIO representation
        path (Path): Path where to save the image

    Returns:
        int: 1 ok, 0 failed
    """
    # FIXME non-existing path testing

    # writing to disk
    f = open(path, "wb+")
    f.write(img_io.getvalue())
    f.close()
    return 1


def normalize_image_extension(extension):
    """Return extensions naming for PIL and Mime/type.

    Args:
        extension (str): extension to normalize.

    Returns:
        list: [PIL extension codename, mimetye]
    """
    web_extension = None
    extension_codename = None

    if extension.lower() == ".jpg" or extension.lower() == ".jpeg":
        extension_codename = "JPEG"
        web_extension = "image/jpeg"

    elif extension.lower() == ".png":
        extension_codename = "PNG"
        web_extension = "image/png"

    elif extension.lower() == ".gif":
        extension_codename = "GIF"
        web_extension = "image/gif"

    elif extension.lower() == ".webp":
        extension_codename = "WEBP"
        web_extension = "image/webp"

    return [extension_codename, web_extension]



class PostProcessor():
    "Plugins that process each post between the parsing and the rendering"

    def process(self, post, site, config):
        """ Process a parsed post to add extra meta or change its HTML.

        Args:
            post (post): post to process.
            site (SiteFab): instanciated site object.
            config (objectdict): plugin configuration.
        """



from sitefab import nlp
from sitefab.nlp import SPACY_MODEL
from textacy import make_spacy_doc


def test_text_cleanup():
    snips = [
                """@test@""",
                " test  ",
                " TeSt ",
                "-#, test, "
    ]

    for snip in snips:
        assert nlp.text_cleanup(snip) == 'test'


def test_stats():
    text = "the quick fox and the cat. The turtle and the rabbit."
    doc = make_spacy_doc(text, lang=SPACY_MODEL)
    stats = nlp.compute_stats(doc)
    assert stats.counts.sentences == 2
    assert stats.counts.words == 11


def test_empyty_stats():
    text = ""
    doc = make_spacy_doc(text, lang=SPACY_MODEL)
    stats = nlp.compute_stats(doc)
    assert stats.counts.sentences == 0
    assert stats.counts.words == 0


def test_terms():
    text = "the quick fox and the cat. The turtle and the rabbit."
    doc = make_spacy_doc(text, lang=SPACY_MODEL)
    terms = nlp.extract_key_terms(doc, num_terms=5)
    terms = [t[0] for t in terms]  # remove scores
    assert 'fox' in terms
    assert 'cat' in terms
    assert 'turtle' in terms
    assert 'rabbit' in terms


def test_empty_terms():
    text = ""
    doc = make_spacy_doc(text, lang=SPACY_MODEL)
    terms = nlp.extract_key_terms(doc, num_terms=5)
    assert terms == []


def test_single_term():
    text = "elie"
    doc = make_spacy_doc(text, lang=SPACY_MODEL)
    terms = nlp.extract_key_terms(doc, num_terms=5)
    assert terms[0][0] == 'elie'
    assert terms[0][1] == 1.0


def test_two_term_behavhior():
    """Case when there is less than 3 words and rank algo can't be used.
    """
    text = "search page"
    doc = make_spacy_doc(text, lang=SPACY_MODEL)
    terms = nlp.extract_key_terms(doc, num_terms=5)
    assert 'search' == terms[0][0]
    assert 0.5 == terms[0][1]
    assert 'page' == terms[1][0]
    assert 0.5 == terms[1][1]


def test_analyze_post(empty_post):
    empty_post.text = """
    Protecting accounts from credential stuffing attacks remains burdensome
    due to an asymmetry of knowledge: attackers have wide-scale access to
    billions of stolen usernames and passwords, while users and identity
    providers remain in the dark as to which accounts require remediation.
    In this paper, we propose a privacy-preserving protocol whereby a
    client can query a centralized breach repository to determine whether
    a specific username and password combination is publicly exposed,
    but without revealing the information queried. Here, a client can be an
    end user, a password manager, or an identity provider. To demonstrate the
    feasibility of our protocol, we implement a cloud service that mediates
    access to over 4 billion credentials found in breaches and a Chrome
    extension serving as an initial client. Based on anonymous telemetry
    from nearly 670,000 users and 21 million logins, we find that 1.5% of
    logins on the web involve breached credentials. By alerting users to this
    breach status, 26% of our warnings result in users migrating to a new
    password, at least as strong as the original. Our study illustrates how
    secure, democratized access to password breach alerting can help
    mitigate one dimension of account hijacking."""

    empty_post.meta.title = 'Protecting accounts from credential stuffing with\
                             password breach alerting'

    post_nlp = nlp.analyze_post(empty_post)
    assert post_nlp.stats.counts.words > 0
    assert post_nlp.stats.readability.flesch_kincaid_grade_level > 0
    assert 'password' in [t[0] for t in post_nlp.title_terms]
    assert 'password' in [t[0] for t in post_nlp.terms]



"""
SiteFab Plugin system
"""

import logging
import os
from pathlib import Path

from termcolor import colored, cprint
from terminaltables import SingleTable
from toposort import toposort_flatten
from tqdm import tqdm
from yapsy.PluginManager import PluginManager

from sitefab import utils

from .CollectionProcessor import CollectionProcessor
from .PostProcessor import PostProcessor
from .SitePreparsing import SitePreparsing
from .SiteProcessor import SiteProcessor
from .SiteRendering import SiteRendering
from .TemplateFilter import TemplateFilter

logging.basicConfig(level=logging.INFO)


class Plugins():
    """
    Class responsible to manage the plugins

    """

    categories = [
        ["SitePreparsing", SitePreparsing,
         "Site wide plugins that execute before the parsing start."],

        ["SiteProcessor", SiteProcessor,
         "Plugins that process the whole site once after parsing."],

        ["SiteRendering", SiteRendering,
         "Plugins that render additional pages after the rendering."],

        ["PostProcessor", PostProcessor,
         "Plugins that process each post after they are parsed"],

        ["CollectionProcessor", CollectionProcessor,
         "Plugins that process each collection after posts are parsed"],

        ["TemplateFilter", TemplateFilter,
         "Plugins that define jinja2 filters to be used in templates"],
    ]

    # for plugin info structure
    PLUGIN_CAT = 0
    PLUGIN_NAME = 1
    PLUGIN_DESC = 2
    PLUGIN_ENABLE = 3
    PLUGIN_MODULE_NAME = 4
    PLUGIN_VERSION = 5

    def __init__(self, plugin_directories, debug_log_fname, plugins_config):
        "Load plugins"

        categories_filter = {}
        for cat in self.categories:
            categories_filter[cat[0]] = cat[1]

        # single or multi-directory handling
        if not isinstance(plugin_directories, list):
            plugin_directories = [plugin_directories]

        self.plugins = PluginManager(plugin_info_ext='sitefab-plugin',
                                     categories_filter=categories_filter)
        self.plugins.setPluginPlaces(plugin_directories)

        self.plugins.locatePlugins()
        self.plugins.loadPlugins()

        self.plugins_config = plugins_config

        # List of enabled plugins
        self.plugins_enabled = {}
        for pl in self.get_plugins_info():
            if pl[self.PLUGIN_ENABLE]:
                self.plugins_enabled[pl[self.PLUGIN_MODULE_NAME]] = 1

        # list of plugins already executed. Used for dependencies tracking
        # across stages
        self.plugins_executed = {}

        # FIXME: make sure it is working
        # logging.basicConfig(filename=debug_log_fname, level=logging.DEBUG)

    def get_plugins(self, category=None):
        """Return the list of plugins

        :param str category: restrict to plugins of a given category

        :rtype: list(iPlugin)
        :return: list of plugins
        """
        if category:
            return self.plugins.getPluginsOfCategory(category)
        else:
            return self.plugins.getAllPlugins()

    def get_num_plugins(self, category=None):
        """ Return the number of plugins available.

        :param str category: restrict to plugins of given category

        :rtype: int
        :return: number of plugins
        """
        plugins = self.get_plugins(category)
        return len(plugins)

    def get_plugin_dir(self, plugin):
        """ Return the directory where the plugin is stored

        :param iPlugin plugin: the plugin requested

        :rtype: str
        :return: the module name
        """

        module_path = Path(plugin.details.get("Core", "module"))
        return module_path.parent

    def get_plugin_default_configuration_filename(self, plugin):
        """ Return the path to the plugin default configuration filename
        """
        try:
            fname = plugin.details.get("Configuration", "Filename")
        except: # noqa
            return ""
        fname = fname.replace('"', '')
        path = self.get_plugin_dir(plugin)
        return path / fname

    def get_plugin_documentation_filename(self, plugin):
        """ Return the path to the plugin documentation
        """
        try:
            fname = plugin.details.get("Documentation", "Filename")
        except: # noqa
            return ""
        path = self.get_plugin_dir(plugin)
        return path / fname

    def get_plugin_class_name(self, plugin):
        """ Return the class of a given plugin

        :param iPlugin plugin: the plugin requested

        :rtype: str
        :return: the module classname
        """
        return plugin.categories[0]

    def get_plugin_module_name(self, plugin):
        """ Return the module name of a given plugin

        :param iPlugin plugin: the plugin requested

        :rtype: str
        :return: the module name
        """
        module_path = plugin.details.get("Core", "module")
        path, filename = os.path.split(module_path)
        return filename

    def get_plugin_config(self, plugin):
        """ Return the configuration of a given plugin

        :param iPlugin plugin: the plugin requested

        :rtype: dict
        :return: plugin configuration
        """
        module_name = self.get_plugin_module_name(plugin)
        class_name = self.get_plugin_class_name(plugin)
        try:
            config = self.plugins_config[class_name][module_name]
        except:  # noqa
            config = {}
        return config

    def get_plugin_dependencies(self, plugin):
        """ Return the dependency of a given plugin

        :param iPlugin plugin: the plugin requested

        :rtype: list
        :return: list of plugins name the plugin depend on
        """

        # No dependencies
        if not plugin.details.has_option("Core", "Dependencies"):
            return set()

        dependencies = set()
        st = plugin.details.get("Core", "Dependencies")
        if "," in st:
            elts = st.split(",")
            for elt in elts:
                dependencies.add(elt.strip())
        else:
            dependencies.add(st)
        return dependencies

    def is_plugin_enabled(self, plugin):
        config = self.get_plugin_config(plugin)
        if config.get('enable'):
            return True
        else:
            return False

    def get_plugins_info(self, category=None):
        """Return the list of plugins available with their type

        :param str category: restrict to plugins of a given category.

        :rtype: list(str)
        :return: list of plugins name
        """
        pl = []
        if category:
            categories = [category]
        else:
            categories = self.plugins.getCategories()
        for cat in categories:
            for plugin in self.plugins.getPluginsOfCategory(cat):
                enabled = self.is_plugin_enabled(plugin)
                module_name = self.get_plugin_module_name(plugin)
                try:
                    version = plugin.version
                except: # noqa
                    version = "NA"

                s = [cat, plugin.name, plugin.description, enabled,
                     module_name, version]
                pl.append(s)
        return pl

    def display_execution_results(self, results, site):
        """ Display execution summary
        """
        cprint("|-Execution result", "magenta")
        table_data = [['name', 'ok', 'skipped', 'errors']]
        count = 0
        for result in results:
            row = []
            plugin_name, stats = result

            # plugin name
            c = "cyan"
            if count % 2:
                c = "blue"
            row.append(colored(plugin_name, c))

            # ok
            if stats[site.OK]:
                if stats[site.OK] == 1:
                    val = 'v'
                else:
                    val = stats[site.OK]
                row.append(colored(val, 'green'))
            else:
                row.append(' ')

            # warning
            if stats[site.SKIPPED]:
                row.append(colored(stats[site.SKIPPED], "yellow"))
            else:
                row.append(' ')

            # error
            if stats[site.ERROR]:
                if stats[site.ERROR] == 1:
                    val = 'x'
                else:
                    val = stats[site.ERROR]
                row.append(colored(val, "red"))
            else:
                row.append(' ')

            table_data.append(row)
            count += 1

        print(SingleTable(table_data).table)

    def run_plugins(self, items, plugin_class, unit, site):
        """Execute a set of plugins on a given list of items

        :param list items: list of items to process
        :param str plugin_type: the plugin_class to use
        :param str unit: the unit to use in the display
        :param SiteFab site: pointer to the site object to be passed
        to the plugins

        :rtype: dict(dict(list))
        :return: plugins execution statistics
        """

        # dependencies map
        dependencie_map = {}

        # used to get back from the module name to the plugin
        module_name_to_plugin = {}

        plugins = self.plugins.getPluginsOfCategory(plugin_class)

        # collecting plugins that are to be executed.
        for plugin in plugins:
            if self.is_plugin_enabled(plugin):
                module_name = self.get_plugin_module_name(plugin)
                module_name_to_plugin[module_name] = plugin

        # dependencies computation.
        # Due to  potential dependencies on plugins from previous stage
        # this must be computed after collecting which
        # plugins were executed.
        for plugin in module_name_to_plugin.values():
            all_dependencies = self.get_plugin_dependencies(plugin)
            dependencies = set()  # topological sort requires use of set
            module_name = self.get_plugin_module_name(plugin)

            for dep_module_name in all_dependencies:
                if dep_module_name not in self.plugins_enabled:
                    utils.error("Plugin:%s can't be executed because\
                                plugin %s is not enable" % (module_name,
                                dep_module_name))

                # only add to the dependencies map the plugins
                # that are from the same stage
                if dep_module_name in module_name_to_plugin:
                    dependencies.add(dep_module_name)
                else:
                    # check if already executed
                    if dep_module_name not in self.plugins_executed:
                        utils.error("Plugin:%s can't be executed because\
                                    plugin %s was not executed in previous\
                                    stage" % (module_name, dep_module_name))

            dependencie_map[module_name] = dependencies

        # print dependencie_map

        # Topological sorting
        try:
            plugins_to_process = toposort_flatten(dependencie_map)
        except Exception as e:
            utils.error("Circular dependencies between plugins.\
                Can't execute plugins:%s" % e)

        s = "|-%s plugins" % (unit.strip().capitalize())
        desc = colored(s, "magenta")
        results = []
        for module_name in tqdm(plugins_to_process, unit=' plugin',
                                desc=desc, leave=True):
            if module_name in module_name_to_plugin:
                plugin = module_name_to_plugin[module_name]
            else:
                raise Exception(
                    "The following plugin module name listed in\
                    dependencies don't exist % s " % module_name)

            pclass = plugin_class.lower()
            filename = "%s.%s.html" % (pclass, module_name)
            log_id = site.logger.create_log(pclass, plugin.name, filename)

            plugin_results = utils.dict_to_objdict({
                site.OK: 0,
                site.SKIPPED: 0,
                site.ERROR: 0
            })

            config = self.get_plugin_config(plugin)

            for item in tqdm(items, unit=unit, desc=plugin.name, leave=False):
                result = plugin.plugin_object.process(item, site, config)
                plugin_results[result[0]] += 1

                severity = result[0]
                name = result[1]
                details = result[2]
                site.logger.record_event(log_id, name, severity, details)

            self.plugins_executed[module_name] = True
            results.append([plugin.name, plugin_results])
            site.logger.write_log(log_id)
        return results

    def get_template_filters(self):
        """Load template filters and return a dictionary list

        Return:
            dict: jinja filter functions
        """
        template_filters = {}

        filters = self.plugins.getPluginsOfCategory("TemplateFilter")

        for flt in filters:
            filter_name = self.get_plugin_module_name(flt)
            template_filters[filter_name] = flt.plugin_object.myfilter

        return template_filters



import sys
import time
from pathlib import Path
from tqdm import tqdm
from collections import defaultdict
from jinja2 import Environment, FileSystemLoader
from termcolor import cprint
from perfcounters import PerfCounters

from sitefab import files
from sitefab.parser import Parser
from sitefab.Logger import Logger
from sitefab.plugins import Plugins
from sitefab.PostCollections import PostCollections
from sitefab.linter.linter import Linter
from sitefab import utils
from sitefab import nlp


class SiteFab(object):
    """ Object representation of the site being built.

    SiteFab main class
    """

    SORT_BY_CREATION_DATE_DESC = 1
    SORT_BY_CREATION_DATE = 2
    SORT_BY_UPDATE_DATE_DESC = 3
    SORT_BY_UPDATE_DATE = 4

    OK = 1
    SKIPPED = 2
    ERROR = 3

    def __init__(self, config_filename, version='1.0'):

        # Timers
        self.cnts = PerfCounters()
        self.cnts.start('Overall')
        self.cnts.start('Init')

        # [configuration]
        self.current_dir = Path.cwd()

        # make the config file path absolute to avoid weird cases
        self.config_filename = Path(config_filename).resolve()
        if not config_filename:
            raise Exception("Supply a configuration filename")

        # exist?
        if self.config_filename.is_file():  # absolute path
            self.config = files.load_config(self.config_filename)
        else:
            utils.error("Config file %s not found" % self.config_filename)

        # site root dir is -1 from where the config is
        self.config.root_dir = self.config_filename.parents[1]
        self.config.build = utils.create_objdict()

        # expose sitefab version to the templates
        self.config.build.sitefab_version = version

        # [parser] #

        # initialize the parser config
        parser_tpl_path = Path(self.config.parser.template_dir)
        self.config.parser.templates_path = (self.config.root_dir /
                                             parser_tpl_path)

        self.config.parser = Parser.make_config(self.config.parser)

        # [plugins]

        # loading configuration
        for d in self.config.plugins.config_dir:
            config_dir = self.config.root_dir / d

            # load the various config files from disk
            plugins_config = defaultdict(dict)
            for config_fname in files.get_files_list(config_dir, '*.yaml'):
                plugin_name = config_fname.stem
                category = "%s%s" % (str(config_fname.parts[-3]).capitalize(),
                                     str(config_fname.parts[-2]).capitalize())
                config = files.load_config(config_fname)
                plugins_config[category][plugin_name] = config

        # where to redirect the standard python log
        debug_log_fname = self.get_logs_dir() / "debug.log"
        self.plugins = Plugins(self.get_plugins_dirs(),
                               debug_log_fname, plugins_config)

        # Store data generated by plugins that can be used later.
        self.plugin_data = {}
        self.plugin_results = defaultdict(int)

        # [template rendering engine] #
        self.jinja2 = Environment(loader=FileSystemLoader(
                                  str(self.get_template_dir())),
                                  extensions=['jinja2.ext.do'])

        # loading templates custom functions
        custom_filters = self.plugins.get_template_filters()
        for flt_name, flt_fct in custom_filters.items():
            self.jinja2.filters[flt_name] = flt_fct

        # [logger] #
        cfg = utils.create_objdict()
        cfg.output_dir = self.get_logs_dir()
        # log template not the one from the users.
        cfg.template_dir = (self.config.root_dir /
                            self.config.logger.template_dir)

        tpl_dir = self.config.root_dir / Path(self.config.logger.template_dir)
        self.config.logger.template_dir = tpl_dir
        cfg.log_template = "log.html"
        cfg.log_index_template = "log_index.html"  # noqa
        cfg.stats_template = "stats.html"
        self.logger = Logger(cfg, self)

        # [linter] #
        linter_config_filename = (self.config.root_dir /
                                  self.config.linter.configuration_file)
        linter_config = files.load_config(linter_config_filename)

        linter_config.report_template_file = (
            self.config.root_dir / self.config.linter.report_template_file)

        linter_config.output_dir = self.get_logs_dir()
        linter_config.site_output_dir = self.get_output_dir()
        self.linter = Linter(linter_config)

        # Finding content and assets.
        self.filenames = utils.create_objdict()
        self.filenames.posts = files.get_files_list(self.get_content_dir(),
                                                    "*.md")

        # Cleanup the output directories.
        files.clean_dir(self.get_output_dir())
        self.cnts.stop('Init')

    def preprocessing(self):
        "Perform pre-processing tasks"
        self.cnts.start('Preprocessing')
        self.execute_plugins([1], "SitePreparsing", " plugin")
        self.cnts.stop('Preprocessing')

    def parse(self):
        "parse md content into post objects"
        self.cnts.start('Parsing')
        filenames = self.filenames.posts
        self.posts = []

        # collections creation
        min_posts = self.config.collections.min_posts

        # posts_by_tag is what is rendered: it contains for as given post both
        #  its tags and its category
        tlp = self.jinja2.get_template(self.config.collections.template)
        path = self.get_output_dir() / self.config.collections.output_dir

        self.posts_by_tag = PostCollections(
            site=self, template=tlp, output_path=path,
            web_path=self.config.collections.output_dir, min_posts=min_posts)

        self.posts_by_category = PostCollections(
            site=self, web_path=self.config.collections.output_dir)

        self.posts_by_template = PostCollections(site=self)

        self.posts_by_microdata = PostCollections(site=self)

        # Parsing
        cprint("\nParsing posts", "magenta")
        progress_bar = tqdm(total=len(filenames),
                            unit=' files', desc="Files", leave=True)
        errors = []
        post_idx = 1
        # NOTE: passing self.config.parser might seems strange
        # but it is required to get the pluging workings
        parser = Parser(self.config.parser, self)
        for filename in filenames:
            file_content = files.read_file(filename)
            post = parser.parse(file_content)
            post.filename = filename
            post.id = post_idx
            post_idx += 1

            # do not process hidden post
            if post.meta.hidden:
                continue

            # Ignore posts set for future date
            if self.config.parser.skip_future and post.meta.creation_date_ts > int(time.time()):
                s = "Post in the future - skipping %s" % (post.meta.title)
                utils.warning(s)
                continue

            # NLP
            post.nlp = nlp.analyze_post(post)

            # Add posts to our list
            self.posts.append(post)

            # insert in template list
            if not post.meta.template:
                errors += "%s has no template defined." % post.filename
                errors += "Can't render it"
                continue

            self.posts_by_template.add(post.meta.template, post)

            # insert in microformat list
            if post.meta.microdata_type:
                self.posts_by_microdata.add(post.meta.microdata_type, post)

            # insert in category
            if post.meta.category:
                self.posts_by_category.add(post.meta.category, post)
                # tags is what is rendered
                self.posts_by_tag.add(post.meta.category, post)

            # insert in tags
            if post.meta.tags:
                for tag in post.meta.tags:
                    self.posts_by_tag.add(tag, post)

            progress_bar.update(1)

        if len(errors):
            utils.error("\n".join(errors))

        self.cnts.stop('Parsing')

    def process(self):
        "Processing stage"

        self.cnts.start('Processing')

        # Posts processing
        print("\nPosts plugins")
        self.execute_plugins(self.posts, "PostProcessor", " posts")

        # Collection processing
        print("\nCollections plugins")
        self.execute_plugins(self.posts_by_category.get_as_list(),
                             "CollectionProcessor", " categories")
        self.execute_plugins(self.posts_by_tag.get_as_list(),
                             "CollectionProcessor", " tags")
        self.execute_plugins(self.posts_by_template.get_as_list(),
                             "CollectionProcessor", " templates")
        self.execute_plugins(self.posts_by_microdata.get_as_list(),
                             "CollectionProcessor", " microdata")

        # site wide processing
        print("\nSite wide plugins")
        self.execute_plugins([1], "SiteProcessor", " site")
        self.cnts.stop('Processing')

    def render(self):
        "Rendering stage"

        self.cnts.start('Rendering')
        print("\nRendering posts")
        self.render_posts()

        print("\nRendering collections")
        self.posts_by_tag.render()

        print("\nAdditional Rendering")
        self.execute_plugins([1], "SiteRendering", " pages")
        self.cnts.stop('Rendering')

    def finale(self):
        "Final stage"

        # Write reminainig logs
        self.logger.write_log_index()
        self.logger.write_stats()

        # Terminal recap
        cprint("Thread used:%s" % self.config.threads, "cyan")

        cprint("\nPerformance", 'magenta')
        self.cnts.stop_all()
        self.cnts.report()

        cprint("Content", 'magenta')
        cprint("|-Num posts: %s" % len(self.posts), "cyan")
        cprint("|-Num categories: %s" %
               self.posts_by_category.get_num_collections(), "yellow")
        cprint("|-Num tags: %s" %
               self.posts_by_tag.get_num_collections(), "cyan")
        cprint("|-Num templates: %s" %
               self.posts_by_template.get_num_collections(), "yellow")

        # PLUGINS
        cprint("\nPlugins", 'magenta')
        cprint("|-Num plugins: %s" % len(self.plugins.plugins_enabled), "cyan")

        if self.plugin_results[self.OK]:
            cprint("|-Num OK:%s " %
                   self.plugin_results[self.OK], 'green')

        if self.plugin_results[self.SKIPPED]:
            cprint("|-Num Skipped:%s " %
                   self.plugin_results[self.SKIPPED], 'yellow')

        if self.plugin_results[self.ERROR]:
            cprint("|-Num Errors:%s" %
                   self.plugin_results[self.ERROR], 'red')

        # LINTER
        cprint("\nLinter", 'magenta')
        self.linter.render_report()  # write the logs

        cprint("|-Posts OK:%s" %
               self.linter.num_post_ok(), 'green')
        cprint("|-Posts with warnings:%s" %
               self.linter.num_post_with_warnings(), 'yellow')
        cprint("|-Posts with errors:%s" % self.linter.num_post_with_errors(),
               'red')

        # Output
        cprint("\nOutput", 'magenta')
        cprint('|-Log index: %s' % (self.get_logs_dir() / 'index.html'),
               'blue')
        cprint('|-Linter log: %s' % (self.get_logs_dir() / 'linter.html'),
               'cyan')
        cprint('|-Generated site: %s' % (self.get_output_dir()), 'yellow')

        cprint('Generation complete - site is ready :)', 'green')

    # Post functions #
    def render_posts(self):
        """Render posts using jinja2 templates."""
        for post in tqdm(self.posts, unit=' pages', miniters=1, desc="Posts"):
            template_name = "%s.html" % post.meta.template
            template = self.jinja2.get_template(template_name)
            rv = template.render(content=post.html, meta=post.meta,
                                 posts=self.posts,
                                 plugin_data=self.plugin_data,
                                 config=self.config,
                                 categories=self.posts_by_category.get_as_dict(),  # noqa: E501
                                 tags=self.posts_by_tag.get_as_dict(),
                                 templates=self.posts_by_template.get_as_dict(),  # noqa: E501
                                 microdata=self.posts_by_microdata.get_as_dict()  # noqa: E501
                                )

            # Linting
            linter_results = self.linter.lint(post, rv, self)
            # Are we stopping on linting errors?
            if linter_results.has_errors and self.config.linter.stop_on_error:
                print(post.filename)
                for err in linter_results.info:
                    print("\t-%s:%s" % (err[0], err[1]))
                sys.exit(-1)

            perm_url = post.meta.permanent_url
            if len(perm_url) and perm_url[0] == '/':
                perm_url = perm_url[1:]
            path = self.get_output_dir() / perm_url
            files.write_file(path, 'index.html', rv)

    # Templates functions #
    def get_num_templates(self):
        "Return the number of templates loaded."
        return len(self.jinja2.list_templates())

    def get_template_list(self):
        "Return the list of templates loaded."
        return self.jinja2.list_templates()

    # Plugins #
    def execute_plugins(self, items, plugin_class, unit):
        """ Execute a given group of plugins

        Args:
            items (str): list of object to apply the plugins to
                         {collection, posts, None}
            plugin_class (str):  which type of plugins to execute
            unit (str): which unit to report in the progress bar

        Return:
            None
        """
        results = self.plugins.run_plugins(items, plugin_class, unit, self)
        self.plugins.display_execution_results(results, self)

        # sum all plugins data for recap
        for result in results:
            plugin, values = result
            for k, v in values.items():
                self.plugin_results[k] += v

    # Files and directories #
    def get_site_info(self):
        "Return site information."
        return self.site

    def get_config(self):
        "Return sitefab config."
        return self.config

    def get_output_dir(self):
        "return the absolute path of the ouput dir"
        return self.config.root_dir / self.config.dir.output

    def get_content_dir(self):
        "return the absolute path of the content dir"
        return self.config.root_dir / self.config.dir.content

    def get_template_dir(self):
        "return the absolute path of the template dir"
        return self.config.root_dir / self.config.dir.templates

    def get_logs_dir(self):
        "return the absolute path of the log dir"
        return self.config.root_dir / self.config.dir.logs

    def get_cache_dir(self):
        "return the absolute path of the cache dir"
        return self.config.root_dir / self.config.dir.cache

    def get_plugins_dirs(self):
        "return the absolute path for the plugins directories"
        dirs = []
        for d in self.config.plugins.dir:
            dirs.append(self.config.root_dir / d)
        return dirs



from collections import Counter

def lint(post, test_info, config, image_info):
    """Check images for potential errors - Error E2xx
    
    Additional test ideas: image height (but already test ratio)
    """
    results = []

    images = post.elements.images
    if not images or not len(images):
        images = []

    # image_info plugin based test
    if image_info != None:
        results += e201_local_img_file_exist(images, test_info, image_info)
        results += e204_banner_width(post, test_info, image_info, config)
        results += e205_image_width(images, test_info, image_info, config)
        results += e206_banner_local_img_file_exist(post, test_info, image_info)
        results += e207_banner_ratio(post, test_info, image_info, config)
    
    results += e202_img_origin(images, test_info, config)
    results += e203_duplicate_image(images, test_info)
    return results

def e201_local_img_file_exist(images, test_info, image_info): # unit_tested: yes
    "Check if all the relative images urls have a coresponding image"
    results = []
    for image in images:
        if image[:4] != "http":
            if not image in image_info:
                results.append(['E201', test_info['E201'] % image])
    return results

def e202_img_origin(images, test_info, config):
    "Check that external images come from an whitelisted source"
    # FIXME lot's of case here for unit tests
    results = []
    for image in images:
        if image[:4] == "http":
            if not config.allowed_image_sources or not len(config.allowed_image_sources):
                results.append(['E202', test_info['E202'] % image])
            else:
                for site in config.allowed_image_sources:
                    if site in image:
                        continue
                    else:
                        results.append(['E202', test_info['E202'] % image]) 
    return results

def e203_duplicate_image(images, test_info): # unit_tested:yes
    "Test if an image is used multiple time"
    results = []
    cnt = Counter(images)
    for img in cnt.most_common():
        if img[1] > 1:
            results.append(['E203', test_info['E203'] % (img[0], img[1])])
    return results

def e204_banner_width(post, test_info, image_info, config):
    "Test if a banner width is above a certain size"
    results = []

    if  not post.meta or not "banner" in post.meta or not post.meta.banner:
        results.append(["no banner in meta", ""])
        return results
    
    if post.meta.banner not in image_info:
        results.append(["banner not in image_info", ""])
        return results

    banner_width = image_info[post.meta.banner]['width']
    if banner_width <  config.min_banner_width:
        results.append(['E204', test_info['E204'] % (post.meta.banner, banner_width)])

    return results

def e205_image_width(images, test_info, image_info, config):
    "Test if the width of images is above a certain size"
    results = []

    for image in images:
        if image in image_info and "width" in image_info[image]:
            width = image_info[image]['width']
            if width <  config.min_image_width:
                results.append(['E205', test_info['E205'] % (image, width)])

    return results

def e206_banner_local_img_file_exist(post, test_info, image_info): # unit_tested: yes
    "Check if banner local file exist"
    results = []
    if post.meta and "banner" in post.meta and post.meta.banner != None and post.meta.banner[:4] != "http":
            if not post.meta.banner in image_info:
                results.append(['E206', test_info['E206'] % post.meta.banner])
    return results

def e207_banner_ratio(post, test_info, image_info, config): # unit_tested: yes
    "Check banner ratio"
    results = []
    expected_ratio = config.banner_size_ratio
    if post.meta and "banner" in post.meta and post.meta.banner[:4] != "http" and post.meta.banner in image_info and expected_ratio > 0:
            ratio = round(float(image_info[post.meta.banner]['width']) / image_info[post.meta.banner]['height'], 1)
            if ratio != expected_ratio:
                results.append(['E207', test_info['E207'] % (post.meta.banner, ratio)])
    return results


""" Handle SiteFab log output
"""
import time
from collections import defaultdict
from collections import Counter
from jinja2 import Environment, FileSystemLoader

from . import utils
from . import files


class Logger():
    """ SiteFab logging system

        Note:
            while the logging system render log in html using jinja2 it use
            a completly separated on to avoid interferring with user
            configuration. Templates are located in the config directory
            under internal_template/
    """
    def __init__(self, config, site):

        self.config = config
        self.site = site  # reference to the main object
        self.logs = {}
        self.jinja2 = Environment(loader=FileSystemLoader(
            str(self.config.template_dir)))
        files.clean_dir(self.config.output_dir)

    # statistics #
    def write_stats(self):
        "Output statistics about the execution"

        # post per category
        cat_stats = Counter()
        cats = self.site.posts_by_category.get_as_dict()
        for tag, data in cats.items():
            cat_stats[tag] = data.meta.num_posts

        # post per tag
        tag_stats = Counter()
        tags = self.site.posts_by_tag.get_as_dict()
        for tag, data in tags.items():
            tag_stats[tag] = data.meta.num_posts

        template = self.jinja2.get_template(self.config.stats_template)
        rv = template.render(cats=cat_stats.most_common(),
                             tags=tag_stats.most_common())
        files.write_file(self.config.output_dir, "stats.html", rv)

    def create_log(self, category, name, filename):
        """ Create a new log
            Usually used to store a plugin output or a phase information
        """
        log = utils.dict_to_objdict()
        log.meta = utils.dict_to_objdict()
        log.events = []

        log.meta.name = name
        log.meta.category = category
        log.meta.filename = filename
        log.meta.start_time = time.time()
        log.meta.num_events = 0
        log.meta.ok = 0
        log.meta.skipped = 0
        log.meta.errors = 0
        log_id = "%s:%s" % (category, name)
        self.logs[log_id] = log
        return log_id

    def record_event(self, log_id, target, severity, details):
        """ Record a event to a given log
        """
        if log_id not in self.logs:
            return False

        # recording event
        event = utils.dict_to_objdict()
        event.time = time.time()
        event.target = target
        event.severity = severity
        event.details = details

        self.logs[log_id].meta.num_events += 1

        # severity
        if severity == self.site.OK:
            self.logs[log_id].meta.ok += 1
            event.severity = "OK"
        elif severity == self.site.SKIPPED:
            event.severity = "SKIPPED"
            self.logs[log_id].meta.skipped += 1
        elif severity == self.site.ERROR:
            event.severity = "ERROR"
            self.logs[log_id].meta.errors += 1
        self.logs[log_id].events.append(event)
        return True

    def write_log(self, log_id):
        """ Write log
        """
        if log_id not in self.logs:
            return False
        lg = self.logs[log_id]
        lg.meta.exec_time = round(time.time() - lg.meta.start_time, 2)
        template = self.jinja2.get_template(str(self.config.log_template))
        rv = template.render(events=lg.events, meta=lg.meta)
        files.write_file(self.config.output_dir, lg.meta.filename, rv)
        return True

    def write_log_index(self):
        " Generate the index.html file that list all generated logs"

        # allows to output by group
        logs = defaultdict(list)
        for l in self.logs.values():
            logs[l.meta.category].append(l)

        template = self.jinja2.get_template(self.config.log_index_template)
        rv = template.render(logs=logs)
        files.write_file(self.config.output_dir, "index.html", rv)



class SiteRendering():
    """Plugins that render additional pages after processing.
    Plugin only called once"""

    def process(self, unused, site, config):
        """ Generate additional page or file after site is rendered

        Args:
            post (post): post to process.
            site (SiteFab): instanciated site object.
            config (objectdict): plugin configuration.
        """



# encoding: utf-8
from .utils import get_linter_errors_list


def test_e104_triggered(sitefab, empty_post):
    empty_post.meta.mylist = ["test", "test"]
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E104" in error_list


def test_e104_not_triggered(sitefab, empty_post):
    empty_post.meta.mylist = ["test"]
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E104" not in error_list


def test_e105_triggered(sitefab, empty_post):
    empty_post.meta.tags = ["tag1", "category", "tag2"]
    empty_post.meta.category = "category"
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E105" in error_list


def test_e105_not_triggered(sitefab, empty_post):
    empty_post.meta.tags = ["tag1", "tag4", "tag2"]
    empty_post.meta.category = "category"
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E105" not in error_list


def test_e106_triggered(sitefab, empty_post):
    empty_post.meta.title = "title with   extra  space"
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E106" in error_list


def test_e106_not_triggered(sitefab, empty_post):
    empty_post.meta.title = "title with no extra space"
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E106" not in error_list


def test_e107_triggered(sitefab, empty_post):
    empty_post.meta.authors = "this is not a list"
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E107" in error_list


def test_e107_not_triggered(sitefab, empty_post):
    empty_post.meta.authors = ['Elie, Bursztein', 'Celine, Bursztein']
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E107" not in error_list


def test_e108_triggered(sitefab, empty_post):
    empty_post.meta.authors = ['Elie, Bursztein', 'No Commas']
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E108" in error_list


def test_e108_not_triggered(sitefab, empty_post):
    empty_post.meta.authors = ['Elie, Bursztein',
                               'Celine, Bursztein', 'John, McBride']
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E108" not in error_list


def test_e109_triggered(sitefab, empty_post):
    empty_post.meta.authors = ['Elie, Bursztein', 'not, Capitalized']
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E109" in error_list


def test_e109_not_triggered(sitefab, empty_post):
    empty_post.meta.authors = ['Elie, Bursztein', 'Celine, Bursztein']
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E109" not in error_list


def test_e110_triggered(sitefab, empty_post):
    empty_post.meta.category = "not Lower"
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E110" in error_list


def test_e110_not_triggered(sitefab, empty_post):
    empty_post.meta.category = "lower"
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E110" not in error_list


def test_e112_triggered(sitefab, empty_post):
    empty_post.meta.files = "not Lower"
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E112" in error_list


def test_e112_not_triggered(sitefab, empty_post):
    empty_post.meta.files = {"pdf": "/test/test.pdf"}
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E112" not in error_list


def test_e113_triggered(sitefab, empty_post):
    values = [53, ['bla'], {'bla': 'oups'}, None]
    for value in values:
        empty_post.meta.banner = value
        results = sitefab.linter.lint(empty_post, "", sitefab)
        error_list = get_linter_errors_list(results)
        assert "E113" in error_list


def test_e113_not_triggered(sitefab, empty_post):
    values = ["test", u"test", "", u""]
    for value in values:
        empty_post.meta.banner = value
        results = sitefab.linter.lint(empty_post, "", sitefab)
        error_list = get_linter_errors_list(results)
        assert "E113" not in error_list


def test_e114_triggered(sitefab, empty_post):
    test_values = [
        "http://",
        "https://"  # empty
        "hTTps://elie.net",  # caps
        "https://elie.net/ test.html"  # space
    ]
    for value in test_values:
        empty_post.meta.banner = value
        results = sitefab.linter.lint(empty_post, "", sitefab)
        error_list = get_linter_errors_list(results)
        assert "E114" in error_list


def test_e114_not_triggered(sitefab, empty_post):
    test_values = [
        "http://www.elie.net",
        "https://www.elie.net",
        "https://www.elie.net/test.html",
        "https://www.elie.net/test/test.html",
        "https://www.elie.net/test/test",
        "https://www.elie.net/test/test.pdf",
        "https://www.elie.net/test/test.pdf?a=42",
        "/cat/file",
        "cat/file",
        "/cat/file.pdf"
        "/cat/file.pdf?a=2"
    ]
    for value in test_values:
        empty_post.meta.banner = value
        results = sitefab.linter.lint(empty_post, "", sitefab)
        error_list = get_linter_errors_list(results)
        assert "E114" not in error_list


def test_e115_triggered(sitefab, empty_post):
    test_values = [
        "file .pdf",
    ]
    for value in test_values:
        empty_post.meta.banner = value
        results = sitefab.linter.lint(empty_post, "", sitefab)
        error_list = get_linter_errors_list(results)
        assert "E115" in error_list


def test_e115_not_triggered(sitefab, empty_post):
    test_values = [
        "file.pdf",
        "/file.pdf",
        "/files/file.pdf",
        "/file/file-test.pdf",
        "/file/file-test02.jpeg",

    ]
    for value in test_values:
        empty_post.meta.banner = value
        results = sitefab.linter.lint(empty_post, "", sitefab)
        error_list = get_linter_errors_list(results)
        assert "E115" not in error_list


def test_e116_triggered(sitefab, empty_post):
    empty_post.meta.title = None
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E116" in error_list


def test_e116_not_triggered(sitefab, empty_post):
    empty_post.meta.title = "this is my title"
    results = sitefab.linter.lint(empty_post, "", sitefab)
    error_list = get_linter_errors_list(results)
    assert "E116" not in error_list


def test_e117_triggered(sitefab, empty_post):
    values = [53, ['bla'], {'bla': 'oups'}, None]
    for value in values:
        empty_post.meta.permanent_url = value
        results = sitefab.linter.lint(empty_post, "", sitefab)
        error_list = get_linter_errors_list(results)
        assert "E117" in error_list


def test_e117_not_triggered(sitefab, empty_post):
    values = ["test", u"test" u"", ""]
    for value in values:
        empty_post.meta.permanent_url = value
        results = sitefab.linter.lint(empty_post, "", sitefab)
        error_list = get_linter_errors_list(results)
        assert "E117" not in error_list


def test_e118_triggered(sitefab, empty_post):
    values = [
        "http://",
        "https://"  # empty
        "hTTps://elie.net",  # caps
        "https://elie.net/ test.html"  # space
    ]
    for value in values:
        empty_post.meta.permanent_url = value
        results = sitefab.linter.lint(empty_post, "", sitefab)
        error_list = get_linter_errors_list(results)
        assert "E118" in error_list


def test_e118_not_triggered(sitefab, empty_post):
    test_values = [
        "/cat/file",
        "/cat/file?a=5"
        "/cat/file.pdf",
        ""
    ]
    for value in test_values:
        empty_post.meta.permanent_url = value
        results = sitefab.linter.lint(empty_post, "", sitefab)
        error_list = get_linter_errors_list(results)
        assert "E118" not in error_list


def test_e119_triggered(sitefab, empty_post):
    values = ["notabsolute", "not/absolute/not"]
    for value in values:
        empty_post.meta.permanent_url = value
        results = sitefab.linter.lint(empty_post, "", sitefab)
        error_list = get_linter_errors_list(results)
        assert "E119" in error_list


def test_e119_not_triggered(sitefab, empty_post):
    values = ["/absolute", "/absolute/not"]
    for value in values:
        empty_post.meta.permanent_url = value
        results = sitefab.linter.lint(empty_post, "", sitefab)
        error_list = get_linter_errors_list(results)
        assert "E119" not in error_list


def test_e120_triggered(sitefab, empty_post):
    import pprint
    empty_post.meta.template = "blog_post"
    values = ["wrong", "wrong/again"]
    for value in values:
        empty_post.meta.permanent_url = value
        pprint.pprint(empty_post)
        results = sitefab.linter.lint(empty_post, "", sitefab)
        error_list = get_linter_errors_list(results)
        assert "E120" in error_list


def test_e120_not_triggered(sitefab, empty_post):
    empty_post.meta.template = "blog_post"
    values = ["/blog/ok", "/blog/ok/as", "/blog/ok/as.well"]
    for value in values:
        empty_post.meta.permanent_url = value
        results = sitefab.linter.lint(empty_post, "", sitefab)
        error_list = get_linter_errors_list(results)
        assert "E120" not in error_list


def test_e121_triggered(sitefab, empty_post):
    values = [{"not a valid type": "myslide"}]
    for value in values:
        empty_post.meta.files = value
        results = sitefab.linter.lint(empty_post, "", sitefab)
        error_list = get_linter_errors_list(results)
        assert "E121" in error_list


def test_e121_not_triggered(sitefab, empty_post):
    values = [{"slides": "myslide"}, {
        "paper": "mypaper"}, {"video": "myvideo"}]
    for value in values:
        empty_post.meta.files = value
        results = sitefab.linter.lint(empty_post, "", sitefab)
        error_list = get_linter_errors_list(results)
        assert "E121" not in error_list


def test_e122_triggered(sitefab, empty_post):
    values = [{"not a valid type": "not a valid file"}]
    for value in values:
        empty_post.meta.files = value
        results = sitefab.linter.lint(empty_post, "", sitefab)
        error_list = get_linter_errors_list(results)
        assert "E122" in error_list


def test_e122_not_triggered(sitefab, empty_post):
    values = [{"slides": "/myslide.pdf"}, {"video": "/files/myvideo.mp4"}]
    for value in values:
        empty_post.meta.files = value
        results = sitefab.linter.lint(empty_post, "", sitefab)
        error_list = get_linter_errors_list(results)
        assert "E122" not in error_list


def test_e123_triggered(sitefab, empty_post):
    values = [{"not a valid type": "/baprefix",
               "video": "files/almostcorrect"}]
    for value in values:
        empty_post.meta.files = value
        results = sitefab.linter.lint(empty_post, "", sitefab)
        error_list = get_linter_errors_list(results)
        assert "E123" in error_list


def test_e123_not_triggered(sitefab, empty_post):
    values = [{"slides": "/files/myslide.pdf"},
              {"video": "/files/myvideo.mp4"}]
    for value in values:
        empty_post.meta.files = value
        results = sitefab.linter.lint(empty_post, "", sitefab)
        error_list = get_linter_errors_list(results)
        assert "E123" not in error_list


def test_e124_triggered(sitefab, empty_post):
    values = [{"not a valid type": "/nosuffix.pdf",
               "video": "files/almostcorrect.pdf"},
              {"video": "htt://whateverr/files/myvideo.mp4"}]
    for value in values:
        empty_post.meta.files = value
        results = sitefab.linter.lint(empty_post, "", sitefab)
        error_list = get_linter_errors_list(results)
        assert "E124" in error_list


def test_e124_not_triggered(sitefab, empty_post):
    values = [{"slides": "/files/myfiles-slides.pdf"},
              {"video": "https://whateverr/files/myvideo.mp4"}]
    for value in values:
        empty_post.meta.files = value
        results = sitefab.linter.lint(empty_post, "", sitefab)
        error_list = get_linter_errors_list(results)
        assert "E124" not in error_list



from collections import Counter, defaultdict

def lint(post, test_info, config):
    "Check the structure of given post for potential errors"
    
    results = []

    if  post.meta == None or "toc" not in post.meta:
        return results
    
    tests = [
        e300_no_h1,
        e301_no_single_headline
    ]

    for test in tests:
        results += test(post, test_info, config)

    return results

def e300_no_h1(post, test_info, config):
    "Check for the absence of H1 in the content of the post"
    results = []
    for item in post.meta.toc:
       if item[1] == 1:
           info = test_info['E300'] % (item[0])
           results.append(['E300', info])
    return results

def e301_no_single_headline(post, test_info, config):
    "Check that there not only one h2, h3, h4"
    results = []
    counter = Counter()
    elts = defaultdict(list)
    for item in post.meta.toc:
        counter[item[1]] += 1
        elts[item[0]]
    
    for val, count in counter.most_common():
        if count == 1:
            info = test_info['E301'] % (val, " ,".join(elts[val]))
            results.append(['E301', info])
    return results


from time import time
from setuptools import setup
from setuptools import find_packages
from setuptools.command.develop import develop
from setuptools.command.install import install
from subprocess import call

long_description = open("README.md").read()
version = '1.0.%s' % int(time())


def install_spacy_model():
    "we need to download the spacy model and rebuild"
    call("python -m spacy download en_core_web_sm".split())
    call("python -m pip install -U spacy-lookups-data".split())


class PostDevelopCommand(develop):
    """Post-installation for development mode."""
    def run(self):
        develop.run(self)
        # post install of spacy model
        install_spacy_model()


class PostInstallCommand(install):
    """Post-installation for installation mode."""
    def run(self):
        install.run(self)
        # post install of spacy model
        install_spacy_model()


setup(
    name='sitefab',
    version=version,
    description='State of the art static website generator for humans',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Elie Bursztein',
    author_email='code@elie.net',
    url='https://github.com/ebursztein/sitefab',
    license='Apache 2',
    entry_points={
        'console_scripts': ['sitefab=sitefab.cmdline.cmdline:main'],
    },
    cmdclass={
            'develop': PostDevelopCommand,
            'install': PostInstallCommand,
            },
    package_data={"": ["*.yaml"]},
    install_requires=[
            'Cython',
            'pyyaml',
            'jinja2',
            'tqdm',
            'termcolor',
            'yapsy',
            'gensim',
            'toposort',
            'Pygments',
            'pytz',
            'diskcache',
            'xxhash',
            'pillow',
            'perfcounters',
            'terminaltables',
            'textacy',
            'spacy',
            'mistune==0.8.4'],
    classifiers=[
            'Development Status :: 5 - Production/Stable',
            'Environment :: Console',
            'Intended Audience :: Developers',
            'Intended Audience :: Education',
            'Intended Audience :: Science/Research',
            'Intended Audience :: System Administrators',
            'License :: OSI Approved :: Apache Software License',
            'Programming Language :: Python :: 3',
            'Operating System :: MacOS',
            'Operating System :: Microsoft :: Windows',
            'Operating System :: POSIX :: Linux',
            'Topic :: Documentation',
            'Topic :: Internet :: WWW/HTTP',
            'Topic :: Internet :: WWW/HTTP :: Site Management'],
    packages=find_packages())



# encoding: utf-8
from collections import Counter
import re
import os

# from https://mathiasbynens.be/demo/url-regex diego's one
VALID_URL = re.compile(r"^(?:(?:https?|ftp):\/\/)(?:\S+(?::\S*)?@)?(?:(?!10(?:\.\d{1,3}){3})(?!127(?:\.\d{1,3}){3})(?!169\.254(?:\.\d{1,3}){2})(?!192\.168(?:\.\d{1,3}){2})(?!172\.(?:1[6-9]|2\d|3[0-1])(?:\.\d{1,3}){2})(?:[1-9]\d?|1\d\d|2[01]\d|22[0-3])(?:\.(?:1?\d{1,2}|2[0-4]\d|25[0-5])){2}(?:\.(?:[1-9]\d?|1\d\d|2[0-4]\d|25[0-4]))|(?:(?:[a-z\\x{00a1}\-\\x{ffff}0-9]+-?)*[a-z\\x{00a1}\-\\x{ffff}0-9]+)(?:\.(?:[a-z\\x{00a1}\-\\x{ffff}0-9]+-?)*[a-z\\x{00a1}\-\\x{ffff}0-9]+)*(?:\.(?:[a-z\\x{00a1}\-\\x{ffff}]{2,})))(?::\d{2,5})?(?:\/[^\s]*)?$")  # noqa

VALID_LOCAL_URL = re.compile(r"^/?[a-z0-9\/_\.\-=\?]+$")

VALID_FILENAME = re.compile(r'^[a-z\/][a-z0-9_\-/\.]+\.[a-z]{1,5}$')


def lint(post, test_info, config):
    "Check the frontmatter of a given post for potential errors"

    results = []

    # Testing is meta exists otherwise bailout
    if not post.meta:
        results.append(['E100', test_info['E100']])
        return results
    # Run metas tests
    tests = [
        e101_mandatory_fields,
        e102_mandatory_fields_for_specific_templates,
        e103_field_value,
        e104_duplicate_value,
        e105_category_in_tags,
        e106_duplicate_spaces,
        e107_e108_e109_authors_formating,
        e110_lowercase_fields,
        e111_e112_local_files_exists,
        e113_e114_e115_banner_properly_formated,
        e116_value_not_null,
        e117_e118_e119_permanent_url_is_properly_formated,
        e120_valid_permanent_url_prefix,
        e121_file_properly_named,
    ]

    for test in tests:
        results += test(post, test_info, config)

    return results


def e101_mandatory_fields(post, test_info, config):
    "Check for the presence of mandatory fields in the meta"
    results = []
    for field in config.frontmatter_mandatory_fields:
        if field not in post.meta:
            results.append(['E101', test_info['E101'] % field])
    return results


def e102_mandatory_fields_for_specific_templates(post, test_info, config):
    "Check for the presense of mandatory field for specific template"
    results = []
    if "template" not in post.meta:
        return results

    if post.meta.template in config.frontmatter_mandatory_fields_by_templates:
        for field in config.frontmatter_mandatory_fields_by_templates[post.meta.template]:  # noqa
            if field not in post.meta:
                info = test_info['E102'] % (field, post.meta.template)
                results.append(['E102', info])
    return results


def e103_field_value(post, test_info, config):
    "Check if the value for specific fields match the list"
    results = []
    for field in config.frontmatter_fields_value:
        if field in post.meta:
            if post.meta[field] not in config.frontmatter_fields_value[field]:
                info = test_info['E103'] % (
                    field, post.meta[field],
                    config.frontmatter_fields_value[field])
                results.append(['E103', info])
    return results


def e104_duplicate_value(post, test_info, config):
    "Check if a value appears twice in a field list"
    results = []
    for field in post.meta:
        value_field = post.meta[field]
        if isinstance(value_field, list):
            count = Counter()
            for elt in value_field:
                try:
                    count[elt] += 1
                except:  # noqa
                    continue

            duplicates = []
            for elt in count.most_common():
                if elt[1] > 1:
                    duplicates.append(elt[0])

            if len(duplicates):
                info = test_info['E104'] % (field, " ,".join(duplicates))
                results.append(['E104', info])
    return results


def e105_category_in_tags(post, test_info, config):
    "Check if the category appears in the tag list"
    results = []
    if "category" in post.meta and "tags" in post.meta:

        if post.meta.tags and post.meta.category in post.meta.tags:
            info = test_info['E105'] % (
                post.meta.category, " ,".join(post.meta.tags))
            results.append(['E105', info])
    return results


def e106_duplicate_spaces(post, test_info, config):
    "Check if there are extra spaces"
    results = []
    for field in post.meta:
        value = post.meta[field]
        if not isinstance(value, list):
            value = [value]
        for elt in value:
            if isinstance(elt, str):
                extra_space = re.search(" {2,}", elt)
                if extra_space:
                    info = test_info['E106'] % (field, elt)
                    results.append(['E106', info])
    return results


def e107_e108_e109_authors_formating(post, test_info, config):
    "Check if the authors list is properly formatted"
    results = []

    if "authors" not in post.meta:
        return results

    authors = post.meta.authors
    if not isinstance(authors, list):
        info = test_info['E107'] % authors
        results.append(['E107', authors])
        return results

    for author in authors:
        if ',' not in author:
            info = test_info['E108'] % authors
            results.append(['E108', info])
        else:
            firstname, lastname = author.replace(' ', '').split(',')
            if not firstname[0].isupper() or not lastname[0].isupper():
                info = test_info['E109'] % firstname
                results.append(['E109', info])
    return results


def e110_lowercase_fields(post, test_info, config):
    "Check that field values are indeed lowercase"
    results = []
    for field in config.frontmatter_field_values_must_be_lowercase:
        if field in post.meta:
            value = post.meta[field]
            if not isinstance(value, list):
                value = [value]
            for elt in value:
                if isinstance(elt, str):
                    if not elt.islower():
                        info = test_info['E110'] % (field, elt)
                        results.append(['E110', info])
    return results


def e111_e112_local_files_exists(post, test_info, config):
    "check if local files exists"
    results = []
    site_dir = config.site_output_dir
    if "files" in post.meta:
        if not isinstance(post.meta.files, dict):
            info = test_info['E112'] % (type(post.meta.files))
            results.append(['E112', info])
            return results

        for fname, fpath in post.meta.files.items():
            if fpath[0] == '/':
                full_path = os.path.join(site_dir, fpath[1:])
                if not os.path.isfile(full_path):
                    info = test_info['E111'] % (fname, full_path)
                    results.append(['E111', info])
    return results


def e113_e114_e115_banner_properly_formated(post, test_info, config):
    "Ensure the banner is properly formated"
    results = []

    if "banner" not in post.meta:
        return results

    banner = post.meta.banner
    if not isinstance(banner, str):
        info = test_info['E113'] % (type(banner))
        results.append(['E113', info])
        return results

    if "http" in banner[:6]:
        if not VALID_URL.match(banner):
            info = test_info['E114'] % (banner)
            results.append(['E114', info])
    else:
        if not VALID_FILENAME.match(banner):
            info = test_info['E115'] % (banner)
            results.append(['E115', info])

    return results


def e116_value_not_null(post, test_info, config):
    "Ensure the field value are not null"
    WHITELIST = ['toc', 'statistics', 'elements']  # those are generated
    results = []
    for field in post.meta:
        if field in WHITELIST:
            continue
        if not post.meta[field]:
            info = test_info['E116'] % (field)
            results.append(['E116', info])
    return results


def e117_e118_e119_permanent_url_is_properly_formated(post, test_info, config):
    results = []

    if "permanent_url" not in post.meta:
        return results

    url = post.meta.permanent_url
    if not isinstance(url, str):
        info = test_info['E117'] % (type(url))
        results.append(['E117', info])
        return results

    if ((url != "" and not VALID_URL.match(url)) and
            (not VALID_LOCAL_URL.match(url))):
        info = test_info['E118'] % (url)
        results.append(['E118', info])

    if len(url) and url[0] != '/':
        info = test_info['E119'] % (url)
        results.append(['E119', info])

    return results


def e120_valid_permanent_url_prefix(post, test_info, config):
    "Check if the permanent url has a valid template based of its prefix"
    results = []

    if "template" not in post.meta or "permanent_url" not in post.meta:
        return results

    tlp = post.meta.template
    if tlp not in config.permanent_url_valid_prefixes_by_template:
        return results

    prefix = config.permanent_url_valid_prefixes_by_template[tlp]
    permanent_url = str(post.meta.permanent_url)
    if not permanent_url.startswith(prefix):
        info = test_info['E120'] % (permanent_url, prefix)
        results.append(['E120', info])

    return results


def e121_file_properly_named(post, test_info, config):
    "Check if the files are properly named"
    results = []
    # test if it contains -slides.pdf or -paper.pdf
    # test it contains the name of the short url (see rename tools)
    if "files" not in post.meta or not isinstance(post.meta.files, dict):
        return results

    for t, f in post.meta.files.items():
        # valid type
        if t not in config.files.valid_types:
            info = test_info['E121'] % (t)
            results.append(['E121', info])

        # valid characters
        if not VALID_URL.match(f) and not VALID_LOCAL_URL.match(f):
            info = test_info['E122'] % (f)
            results.append(['E122', info])

        # valid prefix
        valid = False
        for prefix in config.files.valid_prefixes:
            if f.startswith(prefix):
                valid = True
        if not valid:
            info = test_info['E123'] % (
                f, " ,".join(config.files.valid_prefixes))
            results.append(['E123', info])

        # valid suffix
        valid = False
        for suffix in config.files.valid_suffixes:
            if f.endswith(suffix):
                valid = True
        if not valid and not f.startswith("http"):
            info = test_info['E124'] % (
                f, " ,".join(config.files.valid_suffixes))
            results.append(['E124', info])

    return results



from sitefab.files import read_file
from sitefab.parser import Parser
from ..conftest import TEST_ROOT_DIR


def test_parsing_real_md(sitefab):
    parser = Parser(sitefab.config.parser, sitefab)
    fn = sitefab.config.root_dir / 'content/posts/18-4-of-us-internet-users-got-at-least-one-of-their-account-compromised.md'  # noqa
    md = read_file(fn)
    post = parser.parse(md)
    assert 'internet' in post.html
    assert 'Elie' in post.meta.authors[0]
    assert post.meta.statistics.num_images == 2


def test_parsing_basic_md(sitefab):
    parser = Parser(sitefab.config.parser, sitefab)
    fn = TEST_ROOT_DIR / 'data/basic.md'
    md = read_file(fn)
    post = parser.parse(md)

    # frontmatter
    assert 'Elie, Bursztein' in post.meta.authors
    assert post.meta.title == 'Test'

    # html
    assert 'this is a test.' in post.html
    assert '<h1 id="toc-0">heading 1</h1>' in post.html
    assert '<h2 id="toc-1">heading 2</h2>' in post.html
    assert '<strong>bold test</strong>' in post.html

    # text
    assert 'this is a test' in post.text
    assert '<' not in post.text

    # toc
    # format: name, heading type: {1,2,3,4...}, id
    assert post.meta.toc[0] == ('heading 1', 1, 0)
    assert post.meta.toc[1] == ('heading 2', 2, 1)

    # statistics
    assert post.meta.statistics.num_videos == 1
    assert post.meta.statistics.num_images == 1



import pytest
from sitefab.SiteFab import SiteFab
from.conftest import TEMPLATE_DATA_CONFIG_FILE_PATH


def test_empty_config():
    # is SiteFab raise the correct exception
    with pytest.raises(Exception) as excinfo:
        SiteFab(None)
        assert excinfo.value.message == 'Supply a configuration filename'


def test_non_existing_config():
    # is SiteFab raise the correct exception
    with pytest.raises(Exception) as excinfo:
        SiteFab(None)
        assert 'Configuration file not found:' in excinfo.value.message


def test_valid_config():
    # is SiteFab raise the correct exception
    fname = TEMPLATE_DATA_CONFIG_FILE_PATH
    site = SiteFab(fname)
    assert site.config
    # FIXME add more test for the correctness here.


def test_template_filters(sitefab):
    "ensure custom filters from plugins/template/filters are correctly loaded"
    assert 'str_to_list' in sitefab.jinja2.filters


def test_get_config(sitefab):
    assert sitefab.get_config() == sitefab.config
